#
# Copyright (c) 2015-2018 Canonical, Ltd.
#
# This file is part of Talisker
# (see http://github.com/canonical-ols/talisker).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import *  # noqa

import collections
from datetime import datetime
import functools
from ipaddress import ip_address
import logging
import os
import sys

from werkzeug.wrappers import Request, Response
import talisker
from talisker.util import module_cache, pkg_is_installed
from talisker.render import (
    Content,
    Head,
    Link,
    Table,
    render,
)


__all__ = ['private']
logger = logging.getLogger(__name__)


class TestException(Exception):
    pass


def force_unicode(s):
    if isinstance(s, bytes):
        return s.decode('utf8')
    return s


def info_response(request, title, *content):
    """Return a response rendered using talisker.render."""
    content_type = request.accept_mimetypes.best_match(
        ['text/plain', 'text/html'],
        default='text/plain',
    )
    return Response(
        render(content_type, Head(title), content),
        mimetype=content_type,
    )


PRIVATE_BODY_RESPONSE_TEMPLATE = """
IP address {0} not in trusted network.

REMOTE_ADDR: {1}
X-Forwarded-For: {2}
""".lstrip('\n')


def private(f):
    """Only allow approved source addresses."""

    @functools.wraps(f)
    def wrapper(self, request):
        config = talisker.get_config()
        if not request.access_route:
            # this means something probably bugged in werkzeug, but let's fail
            # gracefully
            return Response('no client ip provided', status='403')

        ip_str = force_unicode(request.access_route[-1])
        ip = ip_address(ip_str)
        if ip.is_loopback or any(ip in network for network in config.networks):
            return f(self, request)
        else:
            msg = PRIVATE_BODY_RESPONSE_TEMPLATE.format(
                ip_str,
                force_unicode(request.remote_addr),
                request.headers.get('x-forwarded-for'))
            return Response(msg, status='403')
    return wrapper


def ok_response():
    return Response(talisker.get_config().revision_id + '\n')


@module_cache
def test_counter():
    # this is created lazily as we may not have prometheus installed.
    # Also, isolate in its own registry. Multiprocess mode doesn't really use
    # registries, but it helps keep things separated, which is useful in
    # testing.
    from prometheus_client import Counter, CollectorRegistry
    return Counter('test', 'test', registry=CollectorRegistry())


class StandardEndpointMiddleware(object):
    """WSGI middleware to provide a standard set of endpoints for a service"""

    urlmap = collections.OrderedDict((
        ('', 'index'),
        ('/index', 'index'),
        ('/check', 'check'),
        ('/metrics', None),
        ('/ping', 'ping'),
        ('/info/config', 'config'),
        ('/info/packages', 'packages'),
        ('/info/workers', None),
        ('/info/logtree', None),
        ('/info/objgraph', None),
        ('/test/sentry', 'error'),
        ('/test/statsd', 'test_statsd'),
        ('/test/prometheus', None),
    ))

    no_index = {'', '/', '/index', '/error'}

    def __init__(self, app, namespace='_status'):
        self.app = app
        self.namespace = namespace
        self.prefix = '/' + namespace
        # Publish /metrics only if prometheus_client is available
        if pkg_is_installed('prometheus-client'):
            self.urlmap['/metrics'] = 'metrics'
            self.urlmap['/test/prometheus'] = 'test_prometheus'
        if pkg_is_installed('logging-tree'):
            self.urlmap['/info/logtree'] = 'logtree'
        if pkg_is_installed('psutil'):
            self.urlmap['/info/workers'] = 'workers'
        if pkg_is_installed('objgraph'):
            self.urlmap['/info/objgraph'] = 'objgraph'

    def __call__(self, environ, start_response):
        request = Request(environ)
        response = None
        if request.path.startswith(self.prefix):
            path = request.path[len(self.prefix):].rstrip('/')
            try:
                funcname = self.urlmap.get(path, None)
                func = getattr(self, funcname)
            except (KeyError, AttributeError, TypeError):
                pass
            else:
                response = func(request)

        if response is None:
            # pass thru to the app
            return self.app(environ, start_response)
        else:
            return response(environ, start_response)

    def index(self, request):
        methods = []
        base = request.host_url.rstrip('/')
        for url, funcname in self.urlmap.items():
            if funcname is None:
                continue
            if url in self.no_index:
                continue
            try:
                func = getattr(self, funcname)
            except AttributeError:
                pass
            else:
                methods.append((
                    Link(url, self.prefix + url, host=base),
                    str(func.__doc__),
                ))
        return info_response(
            request,
            'Status',
            Table(methods),
        )

    def ping(self, request):
        """HAProxy status check"""
        return ok_response()

    def check(self, request):
        """Nagios health check"""
        start = {}

        def nagios_start(status, headers, exc_info=None):
            # save status for inspection
            start['status'] = status
            start['headers'] = headers
            if exc_info:
                start['exc'] = sys.exc_info()

        response = self.app(request.environ, nagios_start)
        if not start:
            # nagios_start has not yet been called
            if isinstance(response, collections.Iterable):
                # force evaluation
                response = b''.join(response)

        if 'exc' in start:
            return Response('error', status=500)
        elif start.get('status', '').startswith('404'):
            # app does not provide /_status/nagios endpoint
            return ok_response()
        else:
            # return app's response
            return Response(response,
                            status=start.get('status', 200),
                            headers=start.get('headers', []))

    @private
    def error(self, request):
        """Raise a TestError for testing"""
        raise TestException('this is a test, ignore')

    @private
    def test_statsd(self, request):
        """Increment statsd metric for testing"""
        statsd = request.environ['statsd']
        statsd.incr('test')
        return Response('Incremented {}.test'.format(statsd._prefix))

    @private
    def test_prometheus(self, request):
        """Increment prometheus metric for testing"""
        if not pkg_is_installed('prometheus-client'):
            return Response('Not Supported', status=501)

        test_counter().inc()
        return Response('Incremented test counter')

    @private
    def metrics(self, request):
        """Endpoint exposing Prometheus metrics"""
        if not pkg_is_installed('prometheus-client'):
            return Response('Not Supported', status=501)

        from prometheus_client import CONTENT_TYPE_LATEST
        from talisker.prometheus import (
            collect_metrics,
            PrometheusLockTimeout,
        )

        try:
            data = collect_metrics()
        except PrometheusLockTimeout:
            msg = 'Failed to acquire prometheus lock to collect metrics'
            logger.exception(msg)
            return Response([msg], status=500, mimetype='text/plain')

        return Response(data, status=200, mimetype=CONTENT_TYPE_LATEST)

    @private
    def logtree(self, request):
        """Display the stdlib logging configuration."""
        import logging_tree
        tree = logging_tree.format.build_description()
        return Response(tree)

    @private
    def config(self, request):
        config = talisker.get_config()
        rows = []
        for name, meta in config.metadata().items():
            if meta.default is None:
                is_default = ''
            else:
                is_default = meta.default == meta.value

            rows.append((
                name,
                meta.value,
                '' if meta.raw is None else repr(meta.raw),
                is_default,
            ))

        return info_response(
            request,
            'Config',
            Content('Config', 'h2'),
            Table(rows, headers=['Name', 'Value', 'Raw Value', 'Is Default'])
        )

    @private
    def packages(self, request):
        """List of python packages installed."""
        import pkg_resources
        rows = []
        for p in sorted(
                pkg_resources.working_set, key=lambda p: p.project_name):
            rows.append((
                p.project_name,
                p._version,
                p.location,
                Link(
                    'PyPI',
                    'https://pypi.org/project/{}/{}/',
                    p.project_name,
                    p._version,
                ),
            ))

        return info_response(
            request,
            'Python Packages',
            Table(
                rows,
                headers=['Package', 'Version', 'Location', 'PyPI Link'],
            )
        )

    @private
    def workers(self, request):
        """Information about workers resource usage."""
        import psutil
        arbiter = psutil.Process(os.getppid())
        workers = arbiter.children()
        workers.sort(key=lambda p: p.pid)

        rows = [format_psutil_row('Gunicorn Master', arbiter)]
        for i, worker in enumerate(workers):
            rows.append(format_psutil_row('Worker {}'.format(i), worker))

        master = arbiter.as_dict(MASTER_FIELDS)
        master['cmdline'] = ' '.join(master['cmdline'])

        environ = master.pop('environ')
        config = talisker.get_config()
        clean_environ = [
            (k, v) for k, v in sorted(environ.items())
            if k in config.METADATA
        ]
        sorted_master = [(k, master[k]) for k in MASTER_FIELDS if k in master]

        return info_response(
            request,
            'Workers',
            Content('Workers', 'h2'),
            Table(rows, headers=HEADERS),
            Content('Process Information', 'h2'),
            Table(sorted_master),
            Content('Process Environment (whitelist)', 'h2'),
            Table(clean_environ),
        )

    @private
    def objgraph(self, request):
        import objgraph
        limit = int(request.args.get('limit', 10))
        types = objgraph.most_common_types(limit=limit, shortnames=False)
        leaking = objgraph.most_common_types(
            limit=limit,
            objects=objgraph.get_leaking_objects(),
            shortnames=False,
        )

        # html only links
        limits = [
            'Number of items: ',
            Link('{}', request.path + '?limit={}', 10).html(),
            Link('{}', request.path + '?limit={}', 20).html(),
            Link('{}', request.path + '?limit={}', 50).html(),
        ]
        return info_response(
            request,
            'Python Objects',
            Content(
                'Python Objects for Worker pid {}'.format(os.getpid()),
                'h1',
            ),
            Content(' '.join(limits), 'p', text=False, escape=False),
            Content('Most Common Objects', 'h2'),
            Table(types),
            Content('Leaking Objects (no referrer)', 'h2'),
            Table(leaking),
        )


MASTER_FIELDS = [
    'cwd',
    'cmdline',
    'exe',
    'uids',
    'gids',
    'terminal',
    'environ',
]

PSUTIL_FIELDS = [
    'pid',
    'create_time',
    'username',
    'nice',
    'memory_percent',
    'memory_full_info',
    'num_fds',
    'open_files',
    'cpu_percent',
    'num_threads',
]

HEADERS = [
    'Name',
    'PID',
    'User',
    'Nice',
    'VSS',
    'RSS',
    'USS',
    'Shared',
    'CPU%',
    'Mem%',
    'Uptime',
    'TCP Conns',
    'Open Files',
    'FDs',
    'Threads',
]


def mb(x):
    return '{}MB'.format(x // (1024 ** 2))


def format_psutil_row(name, process):
    data = process.as_dict(PSUTIL_FIELDS)
    uptime = datetime.now() - datetime.fromtimestamp(data['create_time'])
    # = datetime.fromtimestamp(v).strftime("%Y-%m-%d %H:%M:%S")
    return [
        name,
        data['pid'],
        data['username'],
        data['nice'],
        mb(data['memory_full_info'].vms),
        mb(data['memory_full_info'].rss),
        mb(data['memory_full_info'].uss),
        mb(data['memory_full_info'].shared),
        '{:.1f}%'.format(data['cpu_percent']),
        '{:.1f}%'.format(data['memory_percent']),
        '{:.0f}m'.format(uptime.total_seconds() // 60),
        len(process.connections(kind='tcp')),
        len(data['open_files']),
        data['num_fds'],
        data['num_threads'],
    ]



#
# Copyright (c) 2015-2018 Canonical, Ltd.
#
# This file is part of Talisker
# (see http://github.com/canonical-ols/talisker).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import *  # noqa

from collections import OrderedDict, deque
import logging

from gunicorn.glogging import Logger
from gunicorn.app.wsgiapp import WSGIApplication

import talisker
from talisker.context import CONTEXT
import talisker.logs
import talisker.sentry
import talisker.statsd
import talisker.metrics
import talisker.wsgi


__all__ = [
    'TaliskerApplication',
]

# settings for gunicorn when in development
DEVEL_SETTINGS = {
    'accesslog': '-',
    'timeout': 99999,
    'reload': True,
}


logger = logging.getLogger(__name__)


class GunicornMetric:
    latency = talisker.metrics.Histogram(
        name='gunicorn_latency',
        documentation='Duration of requests served by Gunicorn',
        labelnames=['view', 'status', 'method'],
        statsd='{name}.{view}.{method}.{status}',
        buckets=[4, 8, 16, 32, 64, 128, 256, 512, 1024, 2048, 4096, 8192],
    )

    count = talisker.metrics.Counter(
        name='gunicorn_count',
        documentation='Count of gunicorn requests',
        labelnames=['view', 'status', 'method'],
        statsd='{name}.{view}.{method}.{status}',
    )

    errors = talisker.metrics.Counter(
        name='gunicorn_errors',
        documentation='Count of Gunicorn errors',
        labelnames=['view', 'status', 'method'],
        statsd='{name}.{view}.{method}.{status}',
    )


# We add a synthetic signal, SIGCUSTOM, to gunicorn's known signals. This
# allows us to get gunicorn to process the effects of this signal in the
# arbiter's main loop, rather than within the limited context of the signal
# handlers.  This makes worker clean up serialized and normal python code.

DEAD_WORKERS = deque()  # storage for recording dead pids


def handle_custom():
    """Handler for a fake 'signal', to be called from the arbiter's main loop.

    This performs the aggregation of prometheus metrics files for dead workers,
    in a serialized and safe manner.
    """
    from talisker.prometheus import prometheus_cleanup_worker
    pid = None
    while DEAD_WORKERS:
        try:
            pid = DEAD_WORKERS.popleft()
            logger.info('cleaning up prometheus metrics', extra={'pid': pid})
            prometheus_cleanup_worker(pid)
        except Exception:
            # we should never fail at cleaning up
            logger.exception(
                'failed to cleanup prometheus worker files',
                extra={'pid': pid},
            )
    # Clear any sentry breadcrumbs that might have built up
    # This is not ideal, but it's the only place the master process calls our
    # code, so...
    # If we don't, breadcrumbs keep on collecting forever.
    talisker.clear_contexts()


def gunicorn_on_starting(arbiter):
    """Gunicorn on_starging server hook.

    Sets up the fake signal and handler on the arbiter instance.
    """
    arbiter.SIG_NAMES['SIGCUSTOM'] = 'custom'
    arbiter.handle_custom = handle_custom


def gunicorn_child_exit(server, worker):
    """Gunicorn child_exit server hook.

    Note: this runs in a signal handler context, and thus cannot safely perform
    IO, so no logging :(
    """
    DEAD_WORKERS.append(worker.pid)
    # queue the fake signal for processing
    if 'SIGCUSTOM' not in server.SIG_QUEUE:
        server.SIG_QUEUE.append('SIGCUSTOM')


def gunicorn_pre_request(worker, req):
    """Gunicorn pre_request hook.

    Clear any previous contexts on new request.

    Note: we do this on way in, rather than the way out, to preserve the
    request_id context when there's a timeout.
    """
    talisker.clear_contexts()


class GunicornLogger(Logger):
    """Custom gunicorn logger to use structured logging."""

    def get_response_status(self, resp):
        """Resolve differences in status encoding.

        This can vary based on gunicorn version and worker class."""
        if hasattr(resp, 'status_code'):
            return resp.status_code
        elif isinstance(resp.status, str):
            return int(resp.status[:3])
        else:
            return resp.status

    def get_extra(self, resp, req, environ, request_time, status):
        # the wsgi context has finished by now, so the various bits of relevant
        # information are only in the headers
        headers = dict((k.lower(), v) for k, v in resp.headers)
        extra = OrderedDict()
        extra['method'] = environ.get('REQUEST_METHOD')
        extra['path'] = environ.get('PATH_INFO')
        qs = environ.get('QUERY_STRING')
        if qs is not None:
            extra['qs'] = environ.get('QUERY_STRING')
        extra['status'] = status
        if 'x-view-name' in headers:
            extra['view'] = headers['x-view-name']
        extra['duration_ms'] = round(request_time.total_seconds() * 1000, 3)
        extra['ip'] = environ.get('REMOTE_ADDR', None)
        extra['proto'] = environ.get('SERVER_PROTOCOL')
        extra['length'] = getattr(resp, 'sent', None)
        if 'CONTENT_LENGTH' in environ:
            try:
                extra['request_length'] = int(environ['CONTENT_LENGTH'])
            except ValueError:
                pass
        if 'CONTENT_TYPE' in environ:
            extra['request_type'] = environ['CONTENT_TYPE']
        referrer = environ.get('HTTP_REFERER', None)
        if referrer is not None:
            extra['referrer'] = environ.get('HTTP_REFERER', None)
        if 'HTTP_X_FORWARDED_FOR' in environ:
            extra['forwarded'] = environ['HTTP_X_FORWARDED_FOR']
        extra['ua'] = environ.get('HTTP_USER_AGENT', None)

        msg = "{method} {path}{0}".format('?' if extra['qs'] else '', **extra)
        for name, tracker in getattr(CONTEXT, 'request_tracking', {}).items():
            extra[name + '_count'] = tracker.count
            extra[name + '_time_ms'] = tracker.time

        return msg, extra

    def access(self, resp, req, environ, request_time):
        if not (self.cfg.accesslog or self.cfg.logconfig or self.cfg.syslog):
            return

        status_url = environ.get('PATH_INFO', '').startswith('/_status/')

        if status_url and not talisker.get_config().logstatus:
            return

        status = self.get_response_status(resp)
        msg, extra = self.get_extra(resp, req, environ, request_time, status)

        try:
            self.access_log.info(msg, extra=extra)
        except Exception:
            self.exception()

        if not status_url:
            labels = {
                'view': extra.get('view', 'unknown'),
                'method': extra['method'],
                'status': str(status),
            }

            GunicornMetric.count.inc(**labels)
            if status >= 500:
                GunicornMetric.errors.inc(**labels)
            GunicornMetric.latency.observe(
                extra['duration_ms'], **labels)

    def setup(self, cfg):
        super(GunicornLogger, self).setup(cfg)
        # remove the default error handler, instead let it filter up to root
        self.error_log.propagate = True
        self._set_handler(self.error_log, None, None)
        if cfg.accesslog is not None:
            if cfg.accesslog == '-':
                # just propagate to our root logger
                self.access_log.propagate = True
                self._set_handler(self.access_log, None, None)
            else:
                self._set_handler(
                    self.access_log,
                    cfg.accesslog,
                    fmt=talisker.logs.StructuredFormatter())

    @classmethod
    def install(cls):
        # in case used as a library, rather than via the entrypoint,
        # install the logger globally, as this is the earliest point we can do
        # so, if not using the talisker entry point
        logging.setLoggerClass(talisker.logs.StructuredLogger)


class TaliskerApplication(WSGIApplication):
    def __init__(self, prog, devel=False, debuglog=False):
        self._devel = devel
        self._debuglog = debuglog
        super(TaliskerApplication, self).__init__(prog)

    def load_wsgiapp(self):
        app = super(TaliskerApplication, self).load_wsgiapp()
        app = talisker.wsgi.wrap(app)
        return app

    def init(self, parser, opts, args):
        """Provide talisker specific default config for gunicorn.

        These are just defaults, and can be overridden in cli/config,
        but it is helpful to set them here.
        """
        cfg = super(TaliskerApplication, self).init(parser, opts, args)
        if cfg is None:
            cfg = {}

        cfg['logger_class'] = GunicornLogger
        cfg['pre_request'] = gunicorn_pre_request

        # only enable these if we are doing multiproc cleanup
        if talisker.prometheus_multiproc_cleanup:
            cfg['on_starting'] = gunicorn_on_starting
            cfg['child_exit'] = gunicorn_child_exit

        # development config
        if self._devel:
            logger = logging.getLogger(__name__)
            logger.debug(
                'devel mode: setting gunicorn devel default config',
                extra=DEVEL_SETTINGS)
            cfg.update(DEVEL_SETTINGS)

        return cfg

    def load_config(self):
        super(TaliskerApplication, self).load_config()

        logger = logging.getLogger(__name__)
        # override and warn

        if self.cfg.errorlog != '-':
            logger.warning(
                'ignoring gunicorn errorlog config, talisker logs to stderr',
                extra={'errorlog': self.cfg.errorlog})
            self.cfg.set('errorlog', '-')

        if self.cfg.loglevel.lower() == 'debug' and self._devel:
            # user has configured debug level logging
            self.cfg.set('loglevel', 'DEBUG')
            talisker.logs.enable_debug_log_stderr()

        # ensure gunicorn sends debug level messages when needed
        if self._debuglog:
            self.cfg.set('loglevel', 'DEBUG')

        # override and warn
        if self.cfg.statsd_host or self.cfg.statsd_prefix:
            logger.warning(
                'ignoring gunicorn statsd config, as has no effect when '
                'using talisker, as it uses STATS_DSN env var',
                extra={'statsd_host': self.cfg.statsd_host,
                       'statsd_prefix': self.cfg.statsd_prefix})
            self.cfg.set('statsd_host', None)
            self.cfg.set('statsd_prefix', None)

        # trust but warn
        if self.cfg.logger_class is not GunicornLogger:
            logger.warning(
                'using custom gunicorn logger class - this may break '
                'Talisker\'s logging configuration',
                extra={'logger_class': self.cfg.logger_class})



#
# Copyright (c) 2015-2018 Canonical, Ltd.
#
# This file is part of Talisker
# (see http://github.com/canonical-ols/talisker).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import *  # noqa

from collections import OrderedDict
import logging
import re
import time

import raven
import raven.middleware
import raven.handlers.logging
import raven.breadcrumbs

import talisker
import talisker.request_id
import talisker.logs
from talisker.util import (
    get_rounded_ms,
    module_cache,
    parse_url,
    sanitize_url,
)

__all__ = [
    'get_client',
    'configure_client',
    'set_client',
]

default_processors = set([
    'raven.processors.RemovePostDataProcessor',
    'raven.processors.SanitizePasswordsProcessor',
    'raven.processors.RemoveStackLocalsProcessor',
])

# sql queries and http requests are recorded as explicit breadcrumbs as well as
# logged, so ignore the log breadcrumb
raven.breadcrumbs.ignore_logger('talisker.slowqueries')
raven.breadcrumbs.ignore_logger('talisker.requests')


def clear():
    """Clear any sentry state."""
    raven.context._active_contexts.__dict__.clear()


def ensure_talisker_config(kwargs):
    # ensure default processors
    # this is provided as a list from settings, but we need a set
    # to ensure we don't duplicate
    config = talisker.get_config()
    processors = set(kwargs.get('processors') or [])
    kwargs['processors'] = list(default_processors | processors)

    # override it or it interferes with talisker logging
    if kwargs.get('install_logging_hook'):
        logging.getLogger(__name__).info(
            'ignoring install_logging_hook=True in sentry config '
            '- talisker manages this')
    kwargs['install_logging_hook'] = False

    # flask integration explictly sets options as None
    if kwargs.get('release') is None:
        kwargs['release'] = talisker.get_config().revision_id
    # don't hook libraries by default
    if kwargs.get('hook_libraries') is None:
        kwargs['hook_libraries'] = []

    tags = kwargs.get('tags', {})

    # set from the environment
    if config.unit is not None:
        tags['unit'] = config.unit
    if config.environment is not None:
        tags['environment'] = config.environment
    if config.domain is not None:
        tags['domain'] = config.domain
    kwargs['tags'] = tags

    dsn = kwargs.get('dsn', None)
    if not dsn:
        kwargs['dsn'] = config.sentry_dsn


def log_client(client):
    """Safely log client creation at INFO level."""
    if not client.is_enabled():
        # raven already logs a *disabled* client at INFO level
        return

    # base_url shouldn't have secrets in, but just in case, clean it
    public_dsn = client.remote.get_public_dsn()
    scheme = parse_url(client.remote.base_url).scheme
    url = scheme + ':' + public_dsn
    clean_url = sanitize_url(url)
    msg = 'configured raven DSN'
    extra = {'dsn': clean_url}
    config = talisker.get_config()
    env_cfg = config.raw.get('SENTRY_DSN')
    if env_cfg:
        # make a full url look like a public dsn
        clean_env = sanitize_url(re.sub(r'://(.*):.*@', r'://\1@', env_cfg))
        if clean_env == clean_url:
            msg += ' from SENTRY_DSN config'
            extra['from_env'] = True
        else:
            msg += ' overriding SENTRY_DSN config'
            extra['SENTRY_DSN'] = clean_env
    logging.getLogger(__name__).info(msg, extra=extra)


def add_talisker_context(data):
    rid = talisker.request_id.get()
    if rid:
        data['tags']['request_id'] = rid
    data['extra'].update(talisker.logs.logging_context.flat)

    user = data.get('user')
    if user:
        user.pop('email', None)
        user.pop('username', None)

    breadcrumbs = data.get('breadcrumbs', {}).get('values', [])
    start_time = data['extra'].get('start_time')
    sql_crumbs = []

    for crumb in breadcrumbs:
        if start_time:
            crumb['data']['start'] = get_rounded_ms(
                start_time, crumb['timestamp']
            )
        if crumb['category'] == 'sql':
            sql_crumbs.append(crumb)

    if sql_crumbs:
        summary = sql_summary(sql_crumbs, data['extra'].get('start_time'))
        data['extra']['sql summary'] = summary


def sql_summary(sql_crumbs, start_time):

    def duration(crumb):
        return float(crumb['data'].get('duration', 0))

    sql_crumbs.sort(key=duration, reverse=True)
    sql_time = sum(duration(c) for c in sql_crumbs)
    sql_summary = OrderedDict()
    sql_summary['sql_count'] = len(sql_crumbs)
    sql_summary['sql_time'] = sql_time

    if start_time is not None:
        request_time = (time.time() - start_time) * 1000
        sql_summary['non_sql_time'] = request_time - sql_time
        sql_summary['total_time'] = request_time

    sql_summary['slowest queries'] = [
        OrderedDict([
            ('duration', c['data']['duration']),
            ('query', c['data']['query'])]
        )
        for c in sql_crumbs[:5]
    ]

    return sql_summary


class TaliskerSentryClient(raven.Client):

    @property
    def logger(self):
        return logging.getLogger(__name__)

    @logger.setter
    def logger(self, value):
        """Ignore raven.clients logger"""

    def __init__(self, *args, **kwargs):
        ensure_talisker_config(kwargs)
        super().__init__(*args, **kwargs)

    def build_msg(self, event_type, *args, **kwargs):
        data = super().build_msg(event_type, *args, **kwargs)
        add_talisker_context(data)
        return data

    def set_dsn(self, dsn=None, transport=None):
        super().set_dsn(dsn, transport)
        log_client(self)


class ProxyClientMixin(object):
    """Mixin that overrides self.client to be a property.

    This allows reuse of the various raven handlers, which use a client
    attribute, but have that attribute actually get the global client."""

    @property
    def client(self):
        return get_client()

    @client.setter
    def client(self, client):
        """Ignore, as we should be using the global client."""
        pass


class TaliskerSentryLoggingHandler(
        ProxyClientMixin, raven.handlers.logging.SentryHandler):

    def __init__(self):
        # explicitly pass client, so that the Handler doesn't try create it's
        # own client
        super().__init__(client=get_client())


class TaliskerSentryMiddleware(ProxyClientMixin, raven.middleware.Sentry):

    def __init__(self, application):
        # explicitly pass client, so that the Middleware doesn't try create
        # it's own client
        super().__init__(application, get_client())

    def __call__(self, environ, start_response):
        start_time = time.time()
        environ['start_time'] = start_time
        self.client.extra_context({'start_time': start_time})
        soft_start_timeout = talisker.get_config().soft_request_timeout
        if soft_start_timeout >= 0:

            def soft_timeout_start_response(status, headers, exc_info=None):
                response = start_response(status, headers, exc_info=exc_info)
                duration = (time.time() - environ['start_time']) * 1000
                if (soft_start_timeout is not None
                        and duration > soft_start_timeout):
                    self.client.captureMessage(
                        'Start_response over timeout: {}'
                        .format(soft_start_timeout),
                        level='warning'
                    )
                return response
            return super().__call__(environ, soft_timeout_start_response)
        else:
            return super().__call__(environ, start_response)


_client = None


def get_client(**kwargs):
    global _client
    if _client is None:
        _client = TaliskerSentryClient(**kwargs)
    return _client


def configure_client(**kwargs):
    global _client
    _client = TaliskerSentryClient(**kwargs)
    return _client


def set_client(client):
    global _client
    _client = client
    return _client


# module global so we can add filters to it for celery
@module_cache
def get_log_handler():
    return TaliskerSentryLoggingHandler()



#
# Copyright (c) 2015-2018 Canonical, Ltd.
#
# This file is part of Talisker
# (see http://github.com/canonical-ols/talisker).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import *  # noqa
__metaclass__ = type

import errno
import functools
import logging
import pkg_resources
import sys
import time

import werkzeug.local
from future.moves.urllib.parse import urlparse


# look up table for errno's
# FIXME: maybe add more codes?
ERROR_CODES = errno.errorcode.copy()
ERROR_CODES[-1] = 'EAI_BADFLAGS'
ERROR_CODES[-2] = 'EAI_NONAME'
ERROR_CODES[-3] = 'EAI_AGAIN'
ERROR_CODES[-4] = 'EAI_FAIL'
ERROR_CODES[-5] = 'EAI_NODATA'
ERROR_CODES[-6] = 'EAI_FAMILY'
ERROR_CODES[-7] = 'EAI_SOCKTYPE'
ERROR_CODES[-8] = 'EAI_SERVICE'
ERROR_CODES[-9] = 'EAI_ADDRFAMILY'
ERROR_CODES[-10] = 'EAI_MEMORY'
ERROR_CODES[-11] = 'EAI_SYSTEM'
ERROR_CODES[-12] = 'EAI_OVERFLOW'


def parse_url(url, proto='http'):
    # urlparse won't parse properly without a protocol
    if '://' not in url:
        url = proto + '://' + url
    return urlparse(url)


CLEANED = '{scheme}://{user}{colon}{password}{at}{hostname}{port}{path}{qs}'


def sanitize_url(url):
    """Strips a url of any pw and query strings."""
    parsed = parse_url(url)
    return CLEANED.format(
        scheme=parsed.scheme,
        user=parsed.username or '',
        colon=':' if parsed.password else '',
        password='********' if parsed.password else '',
        at='@' if parsed.username or parsed.password else '',
        hostname=parsed.hostname,
        port=':' + str(parsed.port) if parsed.port else '',
        path=parsed.path,
        qs='?' if parsed.query else '',
    )


def force_unicode(s):
    if isinstance(s, bytes):
        return s.decode('utf8')
    return s


def get_rounded_ms(start_time, now_time=None):
    if now_time is None:
        now_time = time.time()
    ms = (now_time - start_time) * 1000
    return round(ms, 3)


def pkg_is_installed(name):
    try:
        return pkg_resources.get_distribution(name)
    except pkg_resources.DistributionNotFound:
        return False


def pkg_version(name):
    return pkg_resources.get_distribution(name).version


class TaliskerVersionException(Exception):
    pass


def ensure_extra_versions_supported(extra):
    # as this is an optional aid, don't fail if on old version of pip
    try:
        talisker_pkg = pkg_resources.get_distribution('talisker')
        extra_deps = (
            set(talisker_pkg.requires([extra])) - set(talisker_pkg.requires())
        )
        for requirement in extra_deps:
            pkg = pkg_resources.get_distribution(requirement.project_name)
            if pkg.version not in requirement.specifier:
                raise TaliskerVersionException(
                    '{} {} is not supported ({})'.format(
                        requirement.project_name, pkg.version, requirement))
    except Exception:
        logging.getLogger(__name__).debug(
            'skipping ensure_extra_versions_supported as exception occured')

    return True


# module level caches for global objects, means we can store all globals in
# a single place. This is useful when testing, as we can reset globals easily.
_global_cache = {}
_global_dicts = []
_context_locals = []


def module_cache(func):
    """Decorates a function to cache its result in a module dict."""

    # Maybe use id(func) instead? Strings are more debug friendly, though.
    id = func.__module__ + '.' + func.__name__

    @functools.wraps(func)
    def get(*args, **kwargs):
        """Return the object from cache, or create it"""
        if id not in _global_cache:
            _global_cache[id] = func(*args, **kwargs)
        return _global_cache[id]

    @functools.wraps(func)
    def update(*args, **kwargs):
        """Force update of the cached object"""
        _global_cache[id] = func(*args, **kwargs)
        return _global_cache[id]

    def raw_update(item):
        """Set the object in the cache directly"""
        old = _global_cache.get(id, None)
        _global_cache[id] = item
        return old

    # expose the raw function, useful for testing
    get.uncached = func
    get.update = update
    get.raw_update = raw_update

    return get


def module_dict():
    d = {}
    _global_dicts.append(d)
    return d


def clear_globals():
    _global_cache.clear()
    for d in _global_dicts:
        d.clear()


def context_local():
    local = werkzeug.local.Local()
    _context_locals.append(local)
    return local


def clear_context_locals():
    for local in _context_locals:
        werkzeug.local.release_local(local)


if sys.version_info[0:2] >= (3, 3):
    def get_root_exception(exc):
        root = exc
        while root.__cause__ is not None or root.__context__ is not None:
            if root.__cause__ is not None:
                root = root.__cause__
            elif root.__context__ is not None:
                root = root.__context__
        return root

else:
    def get_root_exception(exc):
        return exc


def get_errno_fields(exc):
    """Best effort attempt to get any POSIX errno codes from exception."""
    root = get_root_exception(exc)
    fields = {}
    # these fields are standard fields in the OSError heirarchy
    if getattr(root, 'errno', None):
        fields['errno'] = ERROR_CODES.get(root.errno, str(root.errno))
    if getattr(root, 'strerror', None):
        fields['strerror'] = root.strerror
    if getattr(root, 'filename', None) is not None:
        fields['filename'] = root.filename
    if getattr(root, 'filename2', None) is not None:
        fields['filename2'] = root.filename2
    return fields



#
# Copyright (c) 2015-2018 Canonical, Ltd.
#
# This file is part of Talisker
# (see http://github.com/canonical-ols/talisker).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import *  # noqa

import logging
import time

import talisker
import talisker.logs
import talisker.metrics
import talisker.request_id
from talisker.sentry import ProxyClientMixin, get_log_handler
from talisker.util import module_cache


__all__ = [
    'enable_signals',
]


class CeleryMetric:

    count = talisker.metrics.Counter(
        name='celery_count',
        documentation='Count of all celery jobs',
        labelnames=['job'],
        statsd='{name}.{job}',
    )

    success = talisker.metrics.Counter(
        name='celery_success',
        documentation='Count of celery successful jobs',
        labelnames=['job'],
        statsd='{name}.{job}',
    )

    failure = talisker.metrics.Counter(
        name='celery_failure',
        documentation='Count of celery failed jobs',
        labelnames=['job'],
        statsd='{name}.{job}',
    )

    retry = talisker.metrics.Counter(
        name='celery_retry',
        documentation='Count of celery retried jobs',
        labelnames=['job'],
        statsd='{name}.{job}',
    )

    revoked = talisker.metrics.Counter(
        name='celery_revoke',
        documentation='Count of celery revoked jobs',
        labelnames=['job'],
        statsd='{name}.{job}',
    )

    latency = talisker.metrics.Histogram(
        name='celery_latency_run',
        documentation='Latency of celery jobs run time',
        labelnames=['job'],
        statsd='{name}.{job}',
    )

    enqueue_latency = talisker.metrics.Histogram(
        name='celery_latency_enqueue',
        documentation='Latency of celery jobs queue time',
        labelnames=['job'],
        statsd='{name}.{job}',
    )

    queue_latency = talisker.metrics.Histogram(
        name='celery_latency_queue',
        documentation='Latency of celery jobs wait in the queue',
        labelnames=['job'],
        statsd='{name}.{job}',
    )


def _counter(metric):
    """Create a signal handler that counts metrics"""
    def signal(sender, **kwargs):
        metric.inc(job=sender.name)
    return signal


def _protected_counter(metric):
    """Count metrics, but ensure only once.

    This is needed when tasks are eagerly invoked and have retries, or else
    metrics will be duplicated."""
    attr = '_talisker_sent_' + str(metric)

    def protected_signal(sender, **kwargs):
        if not hasattr(sender, attr):
            metric.inc(job=sender.name)
            setattr(sender, attr, True)
    return protected_signal


task_retry = _counter(CeleryMetric.retry)
task_success = _protected_counter(CeleryMetric.success)
task_failure = _protected_counter(CeleryMetric.failure)
task_revoked = _protected_counter(CeleryMetric.revoked)


REQUEST_ID = 'talisker_request_id'
ENQUEUE_START = 'talisker_enqueue_start'


def get_store(body, headers):
    """celery 3.1/4.0 compatability shim."""
    if isinstance(body, tuple):  # celery 4.0.x
        return headers
    else:  # celery 3.1.x
        return body


def get_header(request, header):
    attr = getattr(request, header, None)
    if attr is not None:
        return attr
    if request.headers:
        return request.headers.get(header)
    return None


def before_task_publish(sender, body, headers, **kwargs):
    get_store(body, headers)[ENQUEUE_START] = time.time()
    rid = talisker.request_id.get()
    if rid is not None:
        headers[REQUEST_ID] = rid
    CeleryMetric.count.inc(job=sender)


def after_task_publish(sender, body, **kwargs):
    start_time = get_store(body, kwargs.get('headers', {})).get(ENQUEUE_START)
    if start_time is not None:
        ms = (time.time() - start_time) * 1000
        CeleryMetric.enqueue_latency.observe(ms, job=sender)


def send_run_metric(name, ts):
    ms = (time.time() - ts) * 1000
    CeleryMetric.latency.observe(ms, job=name)


def task_prerun(sender, task_id, task, **kwargs):
    rid = get_header(task.request, REQUEST_ID)
    if rid is not None:
        talisker.request_id.push(rid)

    start_time = get_header(task.request, ENQUEUE_START)
    if start_time is not None:
        ms = (time.time() - start_time) * 1000
        CeleryMetric.queue_latency.observe(ms, job=sender.name)

    if hasattr(task, 'talisker_timestamp') and task.request.is_eager:
        # eager task, but retry has happened immeadiately, so send metrics
        send_run_metric(sender.name, task.talisker_timestamp)
        task_retry(sender)

    task.talisker_timestamp = time.time()
    talisker.logs.logging_context.push(task_id=task_id, task_name=task.name)


def task_postrun(sender, task_id, task, **kwargs):
    if hasattr(task, 'talisker_timestamp'):
        send_run_metric(sender.name, task.talisker_timestamp)
        del task.talisker_timestamp
    talisker.context.clear()


@module_cache
def get_sentry_handler():
    # Need to defer this import so importing talisker.celery doesn't require
    # celery. This is cached, so usually we're only creating one class.
    from raven.contrib.celery import SentryCeleryHandler

    class TaliskerSentryCeleryHandler(ProxyClientMixin, SentryCeleryHandler):
        pass

    return TaliskerSentryCeleryHandler(None)


# By connecting our own no-op handler, we disable celery's logging
# all together
def celery_logging_handler(**kwargs):
    pass  # pragma: no cover


def enable_signals():
    """Best effort enabling of metrics, logging, sentry signals for celery."""
    try:
        from celery import signals
        from raven.contrib.celery import CeleryFilter
    except ImportError:  # pragma: no cover
        return

    signals.setup_logging.connect(celery_logging_handler)
    signals.before_task_publish.connect(before_task_publish)
    signals.after_task_publish.connect(after_task_publish)
    signals.task_prerun.connect(task_prerun)
    signals.task_postrun.connect(task_postrun)
    signals.task_retry.connect(task_retry)
    signals.task_success.connect(task_success)
    signals.task_failure.connect(task_failure)
    signals.task_revoked.connect(task_revoked)

    # install celery error handler
    get_sentry_handler().install()
    # de-dup celery errors
    log_handler = get_log_handler()
    for filter in log_handler.filters:
        if isinstance(filter, CeleryFilter):
            break
    else:
        log_handler.addFilter(CeleryFilter())

    logging.getLogger(__name__).info('enabled talisker celery signals')


def disable_signals():
    from celery import signals
    get_sentry_handler().uninstall()
    signals.setup_logging.disconnect(celery_logging_handler)
    signals.before_task_publish.disconnect(before_task_publish)
    signals.after_task_publish.disconnect(after_task_publish)
    signals.task_prerun.disconnect(task_prerun)
    signals.task_postrun.disconnect(task_postrun)
    signals.task_retry.disconnect(task_retry)
    signals.task_success.disconnect(task_success)
    signals.task_failure.disconnect(task_failure)
    signals.task_revoked.disconnect(task_revoked)



#
# Copyright (c) 2015-2018 Canonical, Ltd.
#
# This file is part of Talisker
# (see http://github.com/canonical-ols/talisker).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import *  # noqa

from contextlib import contextmanager
import functools
import logging
import time

import talisker.statsd


try:
    import prometheus_client
except ImportError:
    prometheus_client = False

try:
    import statsd
except ImportError:
    statsd = False


logger = logging.getLogger(__name__)


class Metric():
    """Abstraction over prometheus and statsd metrics."""

    def __init__(self, name, *args, **kwargs):
        self.name = name
        self.statsd_template = None

        if statsd:
            self.statsd_template = kwargs.pop('statsd', None)

        if prometheus_client:
            self.prometheus = self.get_type()(name, *args, **kwargs)
        else:
            self.prometheus = None

    # prometheus_client does some odd things, if we define this as a class
    # variable it doesn't work
    @property
    def metric_type(self):
        return None

    def get_type(self):
        if prometheus_client:
            return self.metric_type
        return None

    def get_statsd_name(self, labels):
        name = self.name.replace('_', '.')
        if self.statsd_template:
            try:
                name = self.statsd_template.format(name=name, **labels)
            except Exception:
                pass
        return name


def protect(msg):
    def decorator(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            try:
                f(*args, **kwargs)
            except Exception:
                logger.exception(msg)
        return wrapper
    return decorator


class Histogram(Metric):

    @property
    def metric_type(self):
        return prometheus_client.Histogram

    @protect("Failed to collect histogram metric")
    def observe(self, amount, **labels):
        if self.prometheus:
            if labels:
                self.prometheus.labels(**labels).observe(amount)
            else:
                self.prometheus.observe(amount)

        if self.statsd_template:
            client = talisker.statsd.get_client()
            name = self.get_statsd_name(labels)
            client.timing(name, amount)

    @contextmanager
    def time(self):
        """Measure time in ms."""
        t = time.time()
        yield
        d = time.time() - t
        self.observe(d * 1000)


class Counter(Metric):

    @property
    def metric_type(self):
        return prometheus_client.Counter

    @protect("Failed to increment counter metric")
    def inc(self, amount=1, **labels):
        if self.prometheus:
            if labels:
                self.prometheus.labels(**labels).inc(amount)
            else:
                self.prometheus.inc(amount)

        if self.statsd_template:
            client = talisker.statsd.get_client()
            name = self.get_statsd_name(labels)
            client.incr(name, amount)



#
# Copyright (c) 2015-2018 Canonical, Ltd.
#
# This file is part of Talisker
# (see http://github.com/canonical-ols/talisker).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import *  # noqa

from functools import wraps
import uuid
from contextlib import contextmanager

from werkzeug.datastructures import Headers

from talisker.logs import logging_context


__all__ = [
    'HEADER',
    'get',
    'push',
    'context',
    'decorator',
]

HEADER = 'X-Request-Id'


def generate():
    return str(uuid.uuid4())


def get():
    return logging_context.get('request_id')


def push(id):
    return logging_context.push(request_id=id)


# b/w compat alias
set = push


# provide a nicer ctx manager api
@contextmanager
def context(id):
    with logging_context(request_id=id):
        yield


def decorator(id_func):
    """Decorator to set a thread local request id for function.

    Takes a function that will return the request id from the function params,
    so you can configure it. Cleans up on the way out.
    """
    def wrapper(func):
        @wraps(func)
        def decorator(*args, **kwargs):
            id = id_func(*args, **kwargs)

            if id:
                with context(id):
                    return func(*args, **kwargs)
            else:
                return func(*args, **kwargs)

        return decorator
    return wrapper


class RequestIdMiddleware(object):
    """WSGI middleware to set the request id."""

    def __init__(self, app, header=HEADER):
        self.app = app
        self.header = header
        self.wsgi_header = 'HTTP_' + header.upper().replace('-', '_')

    def __call__(self, environ, start_response):
        if self.wsgi_header not in environ:
            environ[self.wsgi_header] = generate()
        rid = environ[self.wsgi_header]
        # don't worry about popping, as wsgi context is cleared
        logging_context.push(request_id=rid)
        environ['REQUEST_ID'] = rid

        def add_id_header(status, response_headers, exc_info=None):
            headers = Headers(response_headers)
            headers.set(self.header, rid)
            start_response(status, headers, exc_info)

        return self.app(environ, add_id_header)



#
# Copyright (c) 2015-2018 Canonical, Ltd.
#
# This file is part of Talisker
# (see http://github.com/canonical-ols/talisker).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import *  # noqa

import collections
import logging
import time
import shlex

from psycopg2.extensions import cursor, connection

try:
    from sqlparse import format as format_sql
except ImportError:
    def format_sql(sql, *args, **kwargs):
        return sql

import raven.breadcrumbs

import talisker
from talisker.util import get_rounded_ms
from talisker.context import track_request_metric

__all__ = [
    'TaliskerConnection',
    'TaliskerCursor',
    'prettify_sql',
]


FILTERED = '<query filtered>'


def prettify_sql(sql):
    if sql is None:
        return None
    return format_sql(
        sql,
        keyword_case="upper",
        identfier_case="lower",
        strip_comments=False,
        reindent=True,
        indent_tabs=False)


def get_safe_connection_string(conn):
    try:
        try:
            # 2.7+
            params = conn.get_dsn_parameters()
        except AttributeError:
            params = dict(i.split('=') for i in shlex.split(conn.dsn))

        params.setdefault('host', 'localhost')
        return '{user}@{host}:{port}/{dbname}'.format(**params)
    except Exception:
        return 'could not parse dsn'


class TaliskerConnection(connection):
    _logger = None
    _threshold = None

    @property
    def logger(self):
        if self._logger is None:
            self._logger = logging.getLogger('talisker.slowqueries')
        return self._logger

    @property
    def query_threshold(self):
        if self._threshold is None:
            self._threshold = talisker.get_config().slowquery_threshold
        return self._threshold

    def cursor(self, *args, **kwargs):
        kwargs.setdefault('cursor_factory', TaliskerCursor)
        return super().cursor(*args, **kwargs)

    def _format_query(self, query):
        if callable(query):
            query = query()
        query = prettify_sql(query)
        query = FILTERED if query is None else query
        return query

    def _record(self, msg, query, duration):
        track_request_metric('sql', duration)

        if self.query_threshold >= 0 and duration > self.query_threshold:
            extra = collections.OrderedDict()
            extra['trailer'] = self._format_query(query)
            extra['duration_ms'] = duration
            extra['connection'] = get_safe_connection_string(self)
            self.logger.info('slow ' + msg, extra=extra)

        def processor(data):
            data['data']['query'] = self._format_query(query)
            data['data']['duration'] = duration
            data['data']['connection'] = get_safe_connection_string(self)

        breadcrumb = dict(
            message=msg, category='sql', data={}, processor=processor)

        raven.breadcrumbs.record(**breadcrumb)


class TaliskerCursor(cursor):

    def execute(self, query, vars=None):
        timestamp = time.time()
        try:
            return super(TaliskerCursor, self).execute(query, vars)
        finally:
            duration = get_rounded_ms(timestamp)
            if vars is None:
                query = None
            self.connection._record('query', query, duration)

    def callproc(self, procname, vars=None):
        timestamp = time.time()
        try:
            return super(TaliskerCursor, self).callproc(procname, vars)
        finally:
            duration = get_rounded_ms(timestamp)
            # no query parameters, cannot safely record
            self.connection._record(
                'stored proc: {}'.format(procname), None, duration)



#
# Copyright (c) 2015-2018 Canonical, Ltd.
#
# This file is part of Talisker
# (see http://github.com/canonical-ols/talisker).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import *  # noqa

from werkzeug.datastructures import Headers

import talisker.request_id
import talisker.context
import talisker.endpoints
import talisker.statsd
import talisker.requests
import talisker.sentry


__all__ = [
    'set_environ',
    'set_headers',
    'wrap'
]


def set_environ(app, **kwargs):
    def middleware(environ, start_response):
        for key, value in kwargs.items():
            environ[key] = value
        return app(environ, start_response)
    return middleware


def set_headers(app, add_headers):
    """Adds headers to response, overwriting any existing values."""
    def middleware(environ, start_response):
        def custom_start_response(status, response_headers, exc_info=None):
            headers = Headers(response_headers)
            for header, value in add_headers.items():
                headers.set(header, value)
            return start_response(status, headers, exc_info)
        return app(environ, custom_start_response)
    return middleware


def wrap(app):
    if getattr(app, '_talisker_wrapped', False):
        return app

    config = talisker.config.get_config()

    wrapped = app
    # added in reverse order
    wrapped = set_headers(
        wrapped, {'X-VCS-Revision': config.revision_id})
    # expose some standard endpoint
    wrapped = talisker.endpoints.StandardEndpointMiddleware(wrapped)
    # set some standard environ items
    wrapped = set_environ(
        wrapped,
        statsd=talisker.statsd.get_client(),
        requests=talisker.requests.get_session(),
    )
    # add request id info to thread locals
    wrapped = talisker.request_id.RequestIdMiddleware(wrapped)
    wrapped = talisker.sentry.TaliskerSentryMiddleware(wrapped)
    wrapped._talisker_wrapped = True
    wrapped._talisker_original_app = app
    return wrapped



#
# Copyright (c) 2015-2018 Canonical, Ltd.
#
# This file is part of Talisker
# (see http://github.com/canonical-ols/talisker).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import *  # noqa
__metaclass__ = type

import collections
from ipaddress import ip_network
import os
import subprocess
import sys

from past.builtins import execfile

from talisker.util import (
    force_unicode,
    module_cache,
    module_dict,
    sanitize_url,
)


__all__ = ['get_config']


# All valid config
CONFIG_META = collections.OrderedDict()
CONFIG_ALIASES = {'TALISKER_COLOUR': 'TALISKER_COLOR'}
# A cache of calculated config values
CONFIG_CACHE = module_dict()
# Collect any configuration errors
CONFIG_ERRORS = module_dict()


def clear():
    CONFIG_CACHE.clear()
    CONFIG_ERRORS.clear()


@module_cache
def get_config(env=os.environ):
    raw = load_env_config(env)
    return Config(raw)


def force_int(n):
    try:
        return int(n)
    except ValueError:
        raise Exception("'{}' is not a valid integer".format(n))


def config_property(raw_name):
    """A special property for declarative configuration specification.

    It collects all the config metadata together, caches parsing logic, and
    provides some convienience when writing configuration logic functions.
    """
    def decorator(func):
        CONFIG_META[raw_name] = (func.__name__, func.__doc__)

        class _property():
            def __get__(self, obj, cls):
                if obj is None:
                    return self
                if raw_name not in CONFIG_CACHE:
                    try:
                        # Note: also passes in the supplied name of the raw
                        # config value, for DRY.
                        CONFIG_CACHE[raw_name] = func(obj, raw_name)
                    except Exception as e:
                        CONFIG_ERRORS[raw_name] = e
                        CONFIG_CACHE[raw_name] = cls.DEFAULTS.get(raw_name)

                return CONFIG_CACHE[raw_name]

        prop = _property()
        prop.__doc__ = func.__doc__
        return prop

    return decorator


class Config():
    """Talisker specific configuration object.

    It takes a 'raw' dict with the unparsed config values in from os.environ or
    file, as appropriate. It then provides python level attributes to access
    that config, which parse the raw values as appropriate.
    """
    ACTIVE = set(['true', '1', 'yes', 'on'])
    INACTIVE = set(['false', '0', 'no', 'off'])
    DEFAULTS = {
        'DEVEL': False,
        'TALISKER_COLOUR': False,
        'TALISKER_LOGSTATUS': False,
        'TALISKER_SLOWQUERY_THRESHOLD': -1,
        'TALISKER_SOFT_REQUEST_TIMEOUT': -1,
        'TALISKER_NETWORKS': [],
    }

    Metadata = collections.namedtuple(
        'Metadata',
        ['name', 'value', 'raw', 'default', 'doc', 'error']
    )

    METADATA = CONFIG_META
    ERRORS = CONFIG_ERRORS
    SANITIZE_URLS = {'SENTRY_DSN'}

    def __init__(self, raw):
        self.raw = raw

    def __getitem__(self, name):
        """Dict-like lookup of raw values."""
        return self.raw.get(name, self.DEFAULTS.get(name))

    def __setitem__(self, name, value):
        """Dict-like setting of raw values, used for testing"""
        CONFIG_CACHE.pop(name, None)
        self.raw[name] = value

    def metadata(self):
        meta = collections.OrderedDict()
        for raw_name, (attr, doc) in self.METADATA.items():
            value = getattr(self, attr)
            if value:
                if raw_name in self.SANITIZE_URLS:
                    value = sanitize_url(value)
                elif isinstance(value, list):
                    value = ', '.join(str(v) for v in value)
            meta[raw_name] = self.Metadata(
                raw_name,
                value,
                self.raw.get(raw_name),
                self.DEFAULTS.get(raw_name),
                doc,
                CONFIG_ERRORS.get(raw_name),
            )
        return meta

    def is_active(self, name):
        """Is the named raw value truthy?"""
        value = self[name]
        if isinstance(value, str):
            return value.lower() in self.ACTIVE
        else:
            return value

    def is_not_active(self, name):
        """Is the named raw value falsey?"""
        value = self[name]
        if isinstance(value, str):
            return value.lower() in self.INACTIVE
        else:
            return value

    @config_property('TALISKER_CONFIG')
    def config_file(self, raw_name):
        """A path to a python file containing configuration variables.

        Note: this will only have effect when set via environment variable, for
        obvious reasons.
        """
        return self[raw_name]

    @config_property('DEVEL')
    def devel(self, raw_name):
        """Allows coloured logs, warnings, and other development convieniences.

        DEVEL mode enables coloured log output, enables python warnings and,
        for gunicorn, it sets longer timeouts, enables access logs, and auto
        reload by default.
        """
        return self.is_active(raw_name)

    @config_property('TALISKER_COLOUR')
    def colour(self, raw_name):
        """Controls the coloured output of logs. Defaults to on if stdin
        is a tty. Can be set to: 0 (off), 1 (on), or 'simple' for a simpler
        colourscheme. Requires DEVEL mode to be enabled.

        Can also be disabled with TERM=dumb env var.
        """

        # explicit disable
        if not self.devel:
            return False
        if os.environ.get('TERM', '').lower() == 'dumb':
            return False

        # was it explicitly set
        if raw_name in self.raw:
            colour = self.raw.get(raw_name).lower()
            if colour in self.ACTIVE:
                return 'default'
            elif colour in self.INACTIVE:
                return False
            else:
                if colour not in ['default', 'simple']:
                    raise Exception(
                        '{} is not a valid colour scheme'.format(colour)
                    )
                return colour
        else:
            # default behaviour when devel=True
            return 'default' if sys.stderr.isatty() else False

    @config_property('DEBUGLOG')
    def debuglog(self, raw_name):
        """Path to write debug level logs to, which is enabled if path is
        writable.

        Debug logs are rotated every 24h to limit size, with only a single
        24 hour archive is kept.
        """
        log = self[raw_name]
        if log is None:
            return None
        else:
            return str(log)

    @config_property('TALISKER_SLOWQUERY_THRESHOLD')
    def slowquery_threshold(self, raw_name):
        """Set the threshold (in ms) over which SQL queries will be logged.
        Defaults to -1 (off). The queries are sanitised, and thus safe to ship
        to a log aggregator.

        Setting to 0 will log every query, which can be useful in development.
        The queries are sanitized by not including the bind parameter values.
        """
        return force_int(self[raw_name])

    @config_property('TALISKER_SOFT_REQUEST_TIMEOUT')
    def soft_request_timeout(self, raw_name):
        """Set the threshold (in ms) over which WSGI requests will report a
        soft time out to sentry. Defaults to -1 (off).

        A soft timeout is simply a warning-level sentry report for the request.
        The aim is to provide early warning and context for when things exceed
        some limit.
        """
        return force_int(self[raw_name])

    @config_property('TALISKER_LOGSTATUS')
    def logstatus(self, raw_name):
        """Sets whether http requests to /_status/ endpoints are logged in
        the access log or not.  Defaults to false.

        These log lines can add a lot of noise, so they are turned off by
        default."""
        return self.is_active(raw_name)

    @config_property('TALISKER_NETWORKS')
    def networks(self, raw_name):
        """Sets additional CIDR networks that are allowed to access restricted
        /_status/ endpoints. Comma separated list.

        This protection is very basic, and can be easily spoofed by
        X-Forwarded-For headers. You should ensure that your HTTP front end
        server is configuring these correctly before passing on to gunicorn.
        """
        network_tokens = self.raw.get(raw_name, '').split()
        networks = [ip_network(force_unicode(n)) for n in network_tokens]
        return networks

    @config_property('TALISKER_REVISION_ID')
    def revision_id(self, raw_name):
        """Sets the explicit revision of the application. If not set, a best
        effort detection of VCS revision is used.

        This is used to tag sentry reports, as well as returned as a header and
        from /_status/ping.

        The default lookup will try find a version-info.txt file, or git, hg,
        or bzr revno, and finally a setup.py version."""
        if raw_name in self.raw:
            return self.raw[raw_name]
        return get_revision_id()

    @config_property('TALISKER_UNIT')
    def unit(self, raw_name):
        """Sets the instance name for use with sentry reports."""
        return self[raw_name]

    @config_property('TALISKER_ENV')
    def environment(self, raw_name):
        """Sets the deployed environment for use with sentry reports (e.g.
        production, staging).
        """
        return self[raw_name]

    @config_property('TALISKER_DOMAIN')
    def domain(self, raw_name):
        """Sets the site domain name for use with sentry reports."""
        return self[raw_name]

    @config_property('STATSD_DSN')
    def statsd_dsn(self, raw_name):
        """Sets the Statsd DSN string, in the form: udp://host:port/my.prefix

        You can also add the querystring parameter ?maxudpsize=N, to change
        from the default of 512.
        """
        return self[raw_name]

    @config_property('SENTRY_DSN')
    def sentry_dsn(self, raw_name):
        """Sets the sentry DSN, as per usual sentry client configuration.

        See the sentry DSN documentation for more details."""
        return self[raw_name]


def parse_config_file(filename):
    module = {
        "__builtins__": __builtins__,
        "__name__": "__config__",
        "__file__": filename,
        "__doc__": None,
        "__package__": None
    }
    execfile(filename, module, module)
    return module


def load_env_config(env=os.environ):
    """Load talisker config from environment"""

    # process aliases
    for name, alias in CONFIG_ALIASES.items():
        if name not in env and alias in env:
            env[name] = env.pop(alias)

    raw_config = dict()
    file_cfg = {}
    path = env.get('TALISKER_CONFIG')

    if path:
        if os.path.exists(path):
            file_cfg = parse_config_file(path)
        else:
            raise RuntimeError('Config file {} does not exists'.format(path))

    for name in CONFIG_META:
        value = env.get(name, file_cfg.get(name, None))
        if value is not None:
            raw_config[name] = value

    return raw_config


def _run(args):
    return subprocess.check_output(args, stderr=subprocess.PIPE)


def version_info_txt():
    with open('version-info.txt', 'rb') as f:
        return f.read()


def git():
    return _run(['git', 'rev-parse', 'HEAD'])


def bzr():
    return _run(['bzr', 'revno'])


def hg():
    return _run(['hg', 'id', '-i'])


def bzr_version_info():
    from versioninfo import version_info
    return version_info['revno']


def setup_py():
    return subprocess.check_output(
        [sys.executable, 'setup.py', '--version'], stderr=subprocess.STDOUT)


revision_funcs = [
    version_info_txt,
    git,
    bzr,
    bzr_version_info,
    hg,
    setup_py,
]


def get_revision_id():
    for func in revision_funcs:
        try:
            rev = func()
            if rev:
                if hasattr(rev, "decode"):
                    return rev.strip().decode('utf8')
                else:
                    return rev.strip()
        except Exception:
            pass
    return u'unknown'



#
# Copyright (c) 2015-2018 Canonical, Ltd.
#
# This file is part of Talisker
# (see http://github.com/canonical-ols/talisker).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import *  # noqa

import collections
import html
from itertools import chain

__all__ = [
    'render',
    'Content',
    'Head',
    'Link',
    'Table',
]

RENDERERS = {
    'text/html': lambda x: x.html(),
    'text/plain': lambda x: x.text(),
}


def render(type, head, content):
    """Render some Content."""
    renderer = RENDERERS[type]
    yield renderer(head)
    for block in content:
        yield renderer(block)


class Content(object):
    """Default simple content object, which can render as text or html."""
    def __init__(
            self, content, tag=None, attrs=None,
            html=True, text=True, escape=True):
        self.content = content
        self.tag = tag
        self.attrs = attrs if attrs is not None else {}
        self.render_html = html
        self.render_text = text
        self.escape = escape

    def html(self):
        if not self.render_html:
            return ''
        return self.html_content()

    def text(self):
        if not self.render_text:
            return ''
        return self.text_content()

    def html_content(self):
        content = self.content
        if self.escape:
            content = html.escape(self.content)
        if self.tag is None:
            return content

        attrs = [
            '{}="{}"'.format(html.escape(k), html.escape(v, quote=True))
            for k, v in self.attrs.items()
        ]
        return '<{tag}{attrs}>{content}</{tag}>'.format(
            tag=self.tag,
            attrs=(" " + " ".join(attrs)) if attrs else '',
            content=content,
        )

    def text_content(self):
        output = [self.content]
        if self.tag in 'h1 h2 h3 h4 h5 h6':
            output.append('=' * len(output[0]))
        output.append('\n')
        return '\n'.join(output)


class Link(Content):
    """A hyperlink."""
    def __init__(self, text, href, *args, **kwargs):
        self.link_text = text.format(*args, **kwargs)
        self.href = href.format(*args, **kwargs)
        self.host = kwargs.get('host', None)
        attrs = kwargs.get('attrs', {})
        attrs['href'] = self.href

        super_kwargs = {
            'text': kwargs.get('text', True),
            'html': kwargs.get('html', True),
        }
        super().__init__(self.link_text, tag='a', attrs=attrs, **super_kwargs)

    def text_content(self):
        if self.host and '://' not in self.href:
            # relative url, not much use in text, so look for host
            return self.host.rstrip('/') + '/' + self.href.lstrip('/')
        else:
            return self.href


class Table(Content):
    def __init__(self, *args, **kwargs):
        self.headers = kwargs.pop('headers', None)
        super().__init__(*args, **kwargs)

    def html_content(self):
        output = ['<table>']
        if self.headers is not None:
            output.append('<thead>')
            output.append('<tr>')
            for header in self.headers:
                if isinstance(header, Content):
                    h = header.html()
                else:
                    h = html.escape(str(header))
                output.append('<th>{}</th>'.format(h))
            output.append('</tr>')
            output.append('</thead>')
        output.append('<tbody>')
        for row in self.content:
            output.append('<tr>')
            for col in row:
                if isinstance(col, Content):
                    c = col.html()
                else:
                    c = html.escape(str(col))
                output.append('<td>{}</td>'.format(c))
            output.append('</tr>')
        output.append('</tbody>')
        output.append('</table>')
        return '\n'.join(output)

    def text_content(self):
        headers = []
        table = []
        widths = collections.defaultdict(int)
        if self.headers is not None:
            for i, header in enumerate(self.headers):
                if isinstance(header, Content):
                    headers.append(header.text())
                else:
                    headers.append(str(header))
                widths[i] = max(widths[i], len(headers[-1]))

        for row in self.content:
            columns = []
            for i, column in enumerate(row):
                if isinstance(column, Content):
                    columns.append(column.text())
                else:
                    columns.append(str(column))
                widths[i] = max(widths[i], len(columns[-1]))
            table.append(columns)

        border = ['-' * (sum(w + 2 for w in widths.values()) - 2)]
        rows = []
        if self.headers:
            rows.append(self.format_table([headers], widths))

        rows.extend([
            border,
            self.format_table(table, widths),
            border,
            ['\n'],
        ])

        return '\n'.join(chain(*rows))

    def format_table(self, table, widths):
        for row in table:
            output = []
            for i, col in enumerate(row):
                w = widths[i]
                if i < len(row) - 1:
                    w += 2
                output.append('{:{w}}'.format(col, w=w))
            yield ''.join(output)


class Head(Content):
    CDN = '//cdnjs.cloudflare.com/ajax/libs'
    HTML_HEADER = """
    <head>
        <title>Talisker: {title}</title>
        <link rel="stylesheet" href="{cdn}/normalize/8.0.0/normalize.min.css">
        <link rel="stylesheet" href="{cdn}/milligram/1.3.0/milligram.min.css">
        <style>body {{{{ margin: 1em }}}}</style>
    </head>
    """.format(cdn=CDN, title='{title}')

    def __init__(self, title):
        super().__init__(title)
        self.title = title

    def html_content(self):
        return self.HTML_HEADER.format(title=self.title)

    def text_content(self):
        title = 'Talisker: {}'.format(self.title)
        border = '=' * len(title)
        return '\n'.join(['', border, title, border, '\n'])



#
# Copyright (c) 2015-2018 Canonical, Ltd.
#
# This file is part of Talisker
# (see http://github.com/canonical-ols/talisker).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import *  # noqa

from collections import OrderedDict
import logging
import logging.handlers
import numbers
import sys
import time

from talisker.context import ContextStack, track_request_metric
from talisker.util import (
    get_errno_fields,
    get_rounded_ms,
    module_cache,
    module_dict,
)


__all__ = [
    'configure',
    'configure_test_logging',
    'logging_context',
]


logging_globals = module_dict()


def set_global_extra(extra):
    if 'extra' not in logging_globals:
        logging_globals['extra'] = OrderedDict()
    logging_globals['extra'].update(extra)


def reset_logging():
    """Reset logging config"""
    # avoid unclosed file resource warning
    for handler in logging.getLogger().handlers:
        if getattr(handler, '_debug_handler', False):
            handler.stream.close()
    logging.getLogger().handlers = []


NOISY_LOGS = {
    'requests': logging.WARNING,
}


logging_context = ContextStack('logging')

# backwards compat alias
set_logging_context = logging_context.push
extra_logging = logging_context


def add_talisker_handler(level, handler, formatter=None):
    if formatter is None:
        formatter = StructuredFormatter()
    handler.setFormatter(formatter)
    handler.setLevel(level)
    handler._talisker_handler = True
    logging.getLogger().addHandler(handler)


def set_logger_class():
    logging.setLoggerClass(StructuredLogger)
    logging.getLogger().setLevel(logging.NOTSET)


@module_cache
def get_talisker_handler():
    handler = logging.StreamHandler()
    handler._root_talisker = True
    return handler


def configure(config):  # pragma: no cover
    """Configure default logging setup for our services.

    This is basically:
     - log to stderr
     - output hybrid logfmt structured format
     - maybe configure debug logging
    """

    # avoid duplicate logging
    if logging_globals.get('configured'):
        return

    set_logger_class()
    formatter = StructuredFormatter()
    if config.colour:
        formatter = ColouredFormatter(style=config.colour)

    # always INFO to stderr
    add_talisker_handler(logging.INFO, get_talisker_handler(), formatter)

    configure_warnings(config.devel)
    supress_noisy_logs()

    # defer this until logging has been set up
    logger = logging.getLogger(__name__)
    config_extra = {k: v.value for k, v in config.metadata().items() if v.raw}
    if config_extra:
        logger.info('talisker configured', extra=config_extra)
    if config.ERRORS:
        errors = {name: str(err) for name, err in config.ERRORS.items()}
        logger.error('configuration errors', extra=errors)

    if config.debuglog is not None:
        if can_write_to_file(config.debuglog):
            handler = logging.handlers.TimedRotatingFileHandler(
                config.debuglog,
                when='D',
                interval=1,
                backupCount=1,
                delay=True,
                utc=True,
            )
            handler._debug_handler = True
            add_talisker_handler(logging.DEBUG, handler)
            logger.info('enabling debug log', extra={'path': config.debuglog})
        else:
            logger.info('could not enable debug log, could not write to path',
                        extra={'path': config.debuglog})

    # sentry integration
    import talisker.sentry  # defer to avoid logging setup
    sentry_handler = talisker.sentry.get_log_handler()
    add_talisker_handler(logging.ERROR, sentry_handler)

    logging_globals['configured'] = True


def can_write_to_file(path):
    try:
        open(path, 'a').close()
    except Exception:
        return False
    else:
        return True


def supress_noisy_logs():
    """Set some custom log levels on some sub logs"""
    for name, level in NOISY_LOGS.items():
        logger = logging.getLogger(name)
        logger.setLevel(level)


def configure_warnings(enable):
    # never propogate warnings to root
    warnings = logging.getLogger('py.warnings')
    warnings.propagate = False

    if enable:
        warnings.addHandler(logging.StreamHandler())


def configure_test_logging(handler=None):
    """Add a handler (defaults to NullHandler) to root logger.

    Prevents unconfigured logging from erroring, and swallows all logging,
    which is usually what you want for unit tests.  Unit test fixtures can
    still add their own loggers to assert against log messages if needed.
    """
    set_logger_class()
    if handler is None:
        handler = logging.NullHandler()
    add_talisker_handler(logging.NOTSET, handler)
    configure_warnings(True)


def enable_debug_log_stderr():
    """Enables debug logging on stderr

    Checks for devel mode."""
    logger = logging.getLogger(__name__)
    logger.warning('setting stderr logging to DEBUG')
    get_talisker_handler().setLevel(logging.DEBUG)


# our enhanced version of the default raven support for recording breadcrumbs
def record_log_breadcrumb(record):
    # lazy import avoids any raven loggers being initialised early
    from raven import breadcrumbs

    breadcrumb_handler_args = (
        logging.getLogger(record.name),
        record.levelno,
        record.message,
        record.args,
        {
            'extra': getattr(record, '_structured', {}),
            'exc_info': record.exc_info,
            'stack_info': getattr(record, 'stack_info', None)
        },
    )

    for handler in getattr(breadcrumbs, 'special_logging_handlers', []):
        if handler(*breadcrumb_handler_args):
            return

    handler = breadcrumbs.special_logger_handlers.get(record.name)
    if handler is not None and handler(*breadcrumb_handler_args):
        return

    def processor(data):
        metadata = {
            'path': record.pathname,
            'lineno': record.lineno,
        }
        if hasattr(record, 'func'):
            metadata['func'] = record.func
        metadata.update(getattr(record, '_structured', {}))
        data.update({
            'message': record.message,
            'category': record.name,
            'level': record.levelname.lower(),
            'data': metadata,
        })
    breadcrumbs.record(processor=processor)


class StructuredLogger(logging.Logger):
    """A logger that handles passing 'extra' arguments to all logging calls.

    Supports 3 sources of extra structured data:

    1) global extra, designed to be set once at process start/
    2) context extra, designed to be set per request or job, can cleaned up
       afterwards.
    3) per call extra, passed by the log call, as per normal logging
       e.g. log.info('...', extra={...})

    """

    def _log(self, *args, **kwargs):
        t = time.time()
        super()._log(*args, **kwargs)
        track_request_metric('logging', get_rounded_ms(t))

    # sadly, we must subclass and override, rather that use the new
    # setLogRecordFactory() in 3.2+, as that does not pass the extra args
    # through. Also, we need to support python 2.
    def makeRecord(self, name, level, fn, lno, msg, args, exc_info,
                   func=None, extra=None, sinfo=None):
        # at this point we have 3 possible sources of extra kwargs
        # - log call: extra
        # - context : local_context.flat
        # - global  : logging_globals['extra']
        #
        # In case of collisions, we append _ to the end of the name, so no data
        # is lost. The global ones are more important, so take priority - the
        # user supplied keys are the ones renamed if needed
        # Also, the ordering is specific - more specific tags first
        trailer = None
        structured = OrderedDict()

        try:
            context_extra = logging_context.flat
            global_extra = logging_globals.get('extra', {})

            if extra:
                trailer = extra.pop('trailer', None)
                for k, v in extra.items():
                    if k in context_extra or k in global_extra:
                        k = k + '_'
                    structured[k] = v

            for k, v in context_extra.items():
                if k in global_extra:
                    k = k + '_'
                structured[k] = v

            structured.update(global_extra)
        except Exception:
            # ensure unexpected error doesn't break logging completely
            structured = extra

        kwargs = dict(func=func, extra=structured, sinfo=sinfo)
        # python 2 doesn't support sinfo parameter
        if sys.version_info[0] == 2:
            kwargs.pop('sinfo')
        record = super(StructuredLogger, self).makeRecord(
            name, level, fn, lno, msg, args, exc_info, **kwargs)
        # store extra explicitly for StructuredFormatter to use
        record.extra = structured
        record._structured = structured  # b/w compat
        record._trailer = trailer
        return record


class StructuredFormatter(logging.Formatter):
    """Add additional structured data in logfmt style to formatted log.

    Outputs the specified format as normal, and adds any structured data
    available in logfmt. Requires StructuredLogger to work.

    e.g.

    2016-01-13 10:24:07.357Z INFO name "my message" foo=data bar="other data"

    """

    FORMAT = '%(asctime)s.%(msecs)03dZ %(levelname)s %(name)s "%(message)s"'
    DATEFMT = "%Y-%m-%d %H:%M:%S"
    MAX_MSG_SIZE = 1024 * 10
    MAX_KEY_SIZE = 256
    MAX_VALUE_SIZE = 1024
    TRUNCATED = '...'
    TRUNCATED_KEY = '___'  # keys cannot have . in

    # use utc time. No idea why this is not the default.
    converter = time.gmtime

    def __init__(self, fmt=None, datefmt=None):
        if fmt is None:
            fmt = self.FORMAT
        if datefmt is None:
            datefmt = self.DATEFMT
        super(StructuredFormatter, self).__init__(fmt, datefmt)

    def format(self, record):
        """Format message with structured tags and any exception/trailer"""
        try:
            record.message = self.clean_message(record.getMessage())

            # we never want sentry to capture DEBUG logs in its breadcrumbs, as
            # they may be sensitive
            if record.levelno > logging.DEBUG:
                record_log_breadcrumb(record)

            if len(record.message) > self.MAX_MSG_SIZE:
                record.message = (
                    record.message[:self.MAX_MSG_SIZE] + self.TRUNCATED
                )

            # this is verbatim from the parent class in stdlib
            if self.usesTime():
                record.asctime = self.formatTime(record, self.datefmt)
            s = self._fmt % record.__dict__

            # add our structured tags *before* exception info is added
            structured = getattr(record, '_structured', {})
            if record.exc_info and 'errno' not in structured:
                structured.update(get_errno_fields(record.exc_info[1]))
            if structured:
                s += " " + self.logfmt(structured)
            # add talisker trailers
            trailer = getattr(record, '_trailer', None)
            if trailer is not None:
                s += '\n' + str(trailer)

            # this is verbatim from the parent class in stdlib
            if record.exc_info:
                # Cache the traceback text to avoid converting it multiple
                # times (it's constant anyway)
                if not record.exc_text:
                    record.exc_text = self.formatException(record.exc_info)
            if record.exc_text:
                if s[-1:] != "\n":
                    s = s + "\n"
                try:
                    s = s + record.exc_text
                except UnicodeError:  # pragma: no cover
                    # Sometimes filenames have non-ASCII chars, which can lead
                    # to errors when s is Unicode and record.exc_text is str
                    # See issue 8924.
                    # We also use replace for when there are multiple
                    # encodings, e.g. UTF-8 for the filesystem and latin-1
                    # for a script. See issue 13232.
                    s = s + record.exc_text.decode(
                        sys.getfilesystemencoding(), 'replace')
            return s
        except Exception:
            # ensure unexpected error doesn't break logging
            return super().format(record)

    def clean_message(self, s):
        return s.replace('"', '\\"').replace('\n', '\\n')

    def logfmt(self, structured):
        formatted = (
            '{}={}'.format(k, v) for k, v in self.logfmt_atoms(structured))
        return " ".join(formatted)

    def logfmt_atoms(self, structured):
        """Generate logfmt atoms from a dict.

        In the case of a bad key, omit it.
        In the case of a bad value, emit key="???"
        """
        key_errors = []
        for k, v in structured.items():
            key = None
            try:
                if v is not False and not v:
                    continue

                key = self.logfmt_key(k)

                if key is None:
                    key_errors.append(k)
                elif isinstance(v, dict):
                    for k2, v2 in v.items():
                        subkey = self.logfmt_key(k2)
                        if subkey is None:
                            key_errors.append(k2)
                        else:
                            yield (key + '_' + subkey, self.logfmt_value(v2))
                else:
                    yield key, self.logfmt_value(v)

            except Exception:
                if key is None:
                    key_errors.append(k)
                else:
                    yield key, '"???"'

        if key_errors:
            # we embed the keys in the message precisely because they
            # couldn't be parsed as logfmt key
            logger = logging.getLogger(__name__)
            logger.warning('could not parse logfmt keys: ' + str(key_errors))

    def logfmt_key(self, k):
        if isinstance(k, bytes):
            k = k.decode('utf8')

        if isinstance(k, str):
            k = k.replace(' ', '_').replace('.', '_').replace('=', '_')
            k = self.safe_string(k, self.MAX_KEY_SIZE, self.TRUNCATED_KEY)
            # TODO: look at measuring perf of this
            # ' ' and = are replaced because they're are not valid logfmt
            # . is replaced because elasticsearch can't do keys with . in
        elif isinstance(k, bool):
            # need to do this here, as bool are also numbers
            return None
        elif isinstance(k, numbers.Number):
            k = str(k)
        else:
            return None

        return k

    def logfmt_value(self, v):
        """Format a python value for logfmt.

        The output target here is influenced by logstash/elastic search types.

        For strings, ensures unicode, quotes, and truncates as needed.
        For bools, format as json truth values.
        For dicts, truncate.
        For others types (e.g. numbers) leave as is.
        For all else, coerce to string.
        """

        if isinstance(v, bytes):
            v = v.decode('utf8')
        if isinstance(v, str):
            v = self.safe_string(v, self.MAX_VALUE_SIZE, self.TRUNCATED)
            if self.string_needs_quoting(v):
                v = '"' + v + '"'
        elif isinstance(v, bool):
            v = str(v).lower()
        elif isinstance(v, numbers.Number):
            v = str(v)
        else:
            v = '"' + str(type(v)) + '"'

        return v

    def string_needs_quoting(self, v):
        all_numeric = True
        decimal_count = 0
        for c in v:
            if c in ' ="\\':
                return True
            if c not in '0123456789.':
                all_numeric = False
            elif c == '.':
                decimal_count += 1

        if all_numeric and decimal_count <= 1:
            return True

        return False

    def safe_string(self, s, max, truncate_str):
        truncated = False
        s = s.strip()

        # no new lines allowed, so truncate at the first new line
        if '\n' in s:
            s = s.split('\n', 1)[0]
            truncated = True

        # maximum size, to prevent overloading aggregator (accidental or
        # malicious)
        if len(s) > max:
            s = s[:max]
            truncated = True

        s = s.replace('"', '\\"')

        if truncated and truncate_str is not None:
            s = s + truncate_str

        return s


DEFAULT_COLOURS = {
    'logfmt': '2;3;36',     # dim italic teal
    'name': '0;33',         # orange
    'msg': '1;16',          # bold white/black, depending on terminal palette
    'time': '2;34',         # dim dark blue
    'DEBUG': '0;32',        # green
    'INFO': '0;32',         # green
    'WARNING': '0;33',      # orange
    'ERROR': '0;31',        # red
    'CRITICAL': '0;31',     # red
}


COLOUR_SCHEMES = {}
COLOUR_SCHEMES['default'] = DEFAULT_COLOURS

# simple strips italics/bold
COLOUR_SCHEMES['simple'] = DEFAULT_COLOURS.copy()
COLOUR_SCHEMES['simple']['logfmt'] = '0;36'
COLOUR_SCHEMES['simple']['msg'] = '0;37'
COLOUR_SCHEMES['simple']['time'] = '0;34'


class ColouredFormatter(StructuredFormatter):
    """Colourised log formatting"""
    CLEAR = '\x1b[0m'

    def __init__(self, style='default'):
        style = COLOUR_SCHEMES[style]
        self.colours = {k: '\x1b[' + v + 'm' for k, v in style.items()}
        format = (
            '{time}%(asctime)s.%(msecs)03dZ{clear} '
            '%(coloured_levelname)s '
            '{name}%(name)s{clear} '
            '"{msg}%(message)s{clear}"'
        ).format(clear=self.CLEAR, **self.colours)
        super().__init__(fmt=format)

    def format(self, record):
        colour = self.colours[record.levelname]
        record.coloured_levelname = '{colour}{levelname}{clear}'.format(
            colour=colour,
            levelname=record.levelname,
            clear=self.CLEAR,
        )
        return super(ColouredFormatter, self).format(record)

    def logfmt(self, structured):
        logfmt_str = super(ColouredFormatter, self).logfmt(structured)
        return self.colours['logfmt'] + logfmt_str + self.CLEAR



#
# Copyright (c) 2015-2018 Canonical, Ltd.
#
# This file is part of Talisker
# (see http://github.com/canonical-ols/talisker).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import *  # noqa

import logging

import flask
import raven.contrib.flask
import talisker.sentry


__all__ = [
    'sentry',
    'register',
    'TaliskerApp',
]


class FlaskSentry(raven.contrib.flask.Sentry):
    _client_set = False

    @property
    def client(self):
        """Return None if not yet set, so we do actually create the client."""
        if self._client_set:
            return talisker.sentry.get_client()
        else:
            return None

    @client.setter
    def client(self, client):
        """We let the flask extension create the sentry client."""
        if client is not None:
            self._client_set = True
            talisker.sentry.set_client(client)

    def after_request(self, sender, response, *args, **kwargs):
        # override after_request to not clear context and transaction
        if self.last_event_id:
            response.headers['X-Sentry-ID'] = self.last_event_id
        return response


def sentry(app, dsn=None, transport=None, **kwargs):
    """Enable sentry for a flask app, talisker style."""
    # transport is just to support testing, not used in prod
    kwargs['logging'] = False
    kwargs['client_cls'] = talisker.sentry.TaliskerSentryClient
    kwargs['wrap_wsgi'] = False
    logging.getLogger(__name__).info('updating raven config from flask app')
    sentry = FlaskSentry(app, **kwargs)
    # tag sentry reports with the flask app
    sentry.client.tags['flask_app'] = app.name
    return sentry


def add_view_name(response):

    name = flask.request.endpoint

    if name is not None and flask.current_app:
        try:
            if name in flask.current_app.view_functions:
                module = flask.current_app.view_functions[name].__module__
                name = module + '.' + name
        except Exception:
            pass

    if name is None:
        # this is not a critical error, so just debug log it.
        logging.getLogger(__name__).debug('no flask view for {}'.format(
            flask.request.path
        ))
    else:
        response.headers['X-View-Name'] = name

    return response


def setup(app):
    sentry(app)
    app.after_request(add_view_name)


def register(app):
    """Register a flask app with talisker."""
    # override flask default app logger set up
    if hasattr(app, 'logger_name'):
        app.config['LOGGER_HANDLER_POLICY'] = 'never'
        app._logger = logging.getLogger(app.logger_name)
    else:
        # we can just set the logger directly
        app.logger = logging.getLogger('flask.app')
    setup(app)


class TaliskerApp(flask.Flask):

    def __init__(self, app, *args, **kwargs):
        super().__init__(app, *args, **kwargs)
        if hasattr(self, 'logger_name'):
            self.config['LOGGER_HANDLER_POLICY'] = 'never'
            self._logger = logging.getLogger(self.logger_name)
        else:
            self._logger = logging.getLogger('flask.app')
        setup(self)

    @property
    def logger(self):
        return self._logger



#
# Copyright (c) 2015-2018 Canonical, Ltd.
#
# This file is part of Talisker
# (see http://github.com/canonical-ols/talisker).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import *  # noqa

import collections
import functools
import logging
import threading
import warnings

from future.moves.urllib.parse import parse_qsl
from future.utils import text_to_native_str
import raven.breadcrumbs
import requests
import werkzeug.local

from talisker.context import track_request_metric
from talisker import request_id
import talisker.metrics
from talisker.util import (
    context_local,
    get_errno_fields,
    module_dict,
    parse_url,
)

__all__ = [
    'configure',
    'enable_requests_logging',
    'get_session',
    'register_endpoint_name',
]

# wsgi requires native strings
HEADER = text_to_native_str(request_id.HEADER)

STORAGE = threading.local()
STORAGE.sessions = {}
HOSTS = module_dict()
_local = context_local()


def clear():
    if hasattr(STORAGE, 'sessions'):
        STORAGE.sessions.clear()
    HOSTS.clear()
    werkzeug.local.release_local(_local)


# storage for metric url state, as requests design allows for no other way
logger = logging.getLogger('talisker.requests')


class RequestsMetric:
    latency = talisker.metrics.Histogram(
        name='requests_latency',
        documentation='Duration of http calls via requests library',
        labelnames=['host', 'view', 'status'],
        statsd='{name}.{host}.{view}.{status}',
    )

    count = talisker.metrics.Counter(
        name='requests_count',
        documentation='Count of http calls via requests library',
        labelnames=['host', 'view'],
        statsd='{name}.{host}.{view}',
    )

    errors = talisker.metrics.Counter(
        name='requests_errors',
        documentation='Count of errors in responses via requests library',
        labelnames=['host', 'type', 'view', 'status'],
        statsd='{name}.{host}.{type}.{view}.{status}',
    )


def register_endpoint_name(endpoint, name):
    """Register a human friendly name for an IP:PORT address for metrics."""
    parsed = parse_url(endpoint)
    HOSTS[parsed.netloc] = name


# backwards compat alias
def register_ip(ip, name):
    warnings.warn(
        "Please use register_endpoint_name", warnings.DeprecationWarning)
    register_endpoint_name(ip, name)


def get_endpoint_name(endpoint):
    parsed = parse_url(endpoint)
    return HOSTS.get(parsed.netloc)


def get_session(cls=requests.Session):
    if not hasattr(STORAGE, 'sessions'):
        STORAGE.sessions = {}
    session = STORAGE.sessions.get(cls)
    if session is None:
        session = STORAGE.sessions[cls] = cls()
        configure(session)
    return session


def configure(session):
    # insert metrics hook first as it needs response, and does't
    # alter hook_data for later hooks
    if metrics_response_hook not in session.hooks['response']:
        session.hooks['response'].insert(0, metrics_response_hook)
    # for some reason, requests doesn't have a pre_request hook anymore
    # so, we do something horrible - we decorate some *instance* methods of the
    # session. This allows us to support injecting request id into the request,
    # and allow a good api for customising the emitted metrics. But it does not
    # require a particular subclass of Session, leaving the user free to use
    # whatever, but still use talisker's enhancements.
    # If requests ever regains request hooks, then maybe this can go away
    # If anyone has a better solution, *please* tell me!
    if not hasattr(session.send, '_send_wrapper'):
        session.send = send_wrapper(session.send)
    if not hasattr(session.request, '_request_wrapper'):
        session.request = request_wrapper(session.request)


def send_wrapper(func):
    """Sets header and records exception details."""
    @functools.wraps(func)
    def send(request, **kwargs):
        rid = request_id.get()
        if rid and HEADER not in request.headers:
            request.headers[HEADER] = rid
        try:
            return func(request, **kwargs)
        except Exception as e:
            record_request(request, None, e)
            raise

    send._send_wrapper = True
    return send


def request_wrapper(func):
    """Adds support for metric_name kwarg to session."""
    @functools.wraps(func)
    def request(method, url, **kwargs):
        try:
            _local.metric_api_name = kwargs.pop('metric_api_name', None)
            _local.metric_host_name = kwargs.pop('metric_host_name', None)
            return func(method, url, **kwargs)
        finally:
            del _local.metric_api_name
            del _local.metric_host_name
    request._request_wrapper = True
    return request


def collect_metadata(request, response):
    metadata = collections.OrderedDict()

    parsed = parse_url(request.url)

    hostname = get_endpoint_name(request.url)
    if hostname is None:
        address = parsed.netloc
        hostname = parsed.hostname
    else:
        address = hostname
        if parsed.port:
            address += ':{}'.format(parsed.port)

    # do not include querystring in url, as may have senstive info
    metadata['url'] = '{}://{}{}'.format(parsed.scheme, address, parsed.path)
    if parsed.query:
        metadata['url'] += '?'
        redacted = ('{}=<len {}>'.format(k, len(v))
                    for k, v in parse_qsl(parsed.query))
        metadata['qs'] = '?' + '&'.join(redacted)
        metadata['qs_size'] = len(parsed.query)

    metadata['method'] = request.method
    metadata['host'] = hostname
    if parsed.netloc not in metadata['url']:
        metadata['netloc'] = parsed.netloc

    if response is not None:
        metadata['status_code'] = response.status_code

        if 'X-View-Name' in response.headers:
            metadata['view'] = response.headers['X-View-Name']
        if 'Server' in response.headers:
            metadata['server'] = response.headers['Server']
        duration = response.elapsed.total_seconds() * 1000
        metadata['duration_ms'] = round(duration, 3)

    request_type = request.headers.get('content-type', None)
    if request_type is not None:
        metadata['request_type'] = request_type

    if metadata['method'] in ('POST', 'PUT', 'PATCH'):
        try:
            metadata['request_size'] = int(
                request.headers.get('content-length', 0))
        except ValueError:
            pass

    if response is not None:
        response_type = response.headers.get('content-type', None)
        if response_type is not None:
            metadata['response_type'] = response_type
        try:
            metadata['response_size'] = int(
                response.headers.get('content-length', 0))
        except ValueError:
            pass

    return metadata


def metrics_response_hook(response, **kwargs):
    """Response hook that records statsd metrics and breadcrumbs."""
    try:
        record_request(response.request, response)
    except Exception:
        logging.exception('response hook error')


def record_request(request, response=None, exc=None):
    metadata = collect_metadata(request, response)
    if response:
        track_request_metric('http', metadata['duration_ms'])

    if exc:
        metadata.update(get_errno_fields(exc))

    raven.breadcrumbs.record(
        type='http',
        category='requests',
        data=metadata,
    )

    labels = {
        'host': metadata['host'],
        'view': metadata.get('view', 'unknown'),
    }

    metric_api_name = getattr(_local, 'metric_api_name', None)
    metric_host_name = getattr(_local, 'metric_host_name', None)
    if metric_api_name is not None:
        labels['view'] = metric_api_name
    if metric_host_name is not None:
        labels['host'] = metric_host_name
    labels['host'] = labels['host'].replace('.', '-')

    RequestsMetric.count.inc(**labels)

    if response is None:
        # likely connection errors
        logger.exception('http request failure', extra=metadata)
        labels['type'] = 'connection'
        labels['status'] = metadata.get('errno', 'unknown')
        RequestsMetric.errors.inc(**labels)
    else:
        logger.info('http request', extra=metadata)
        labels['status'] = metadata['status_code']
        RequestsMetric.latency.observe(metadata['duration_ms'], **labels)
        if metadata['status_code'] >= 500:
            labels['type'] = 'http'
            RequestsMetric.errors.inc(**labels)


def enable_requests_logging():  # pragma: nocover
    """Full requests debug output is tricky to enable"""
    from http.client import HTTPConnection
    HTTPConnection.debuglevel = 1
    logging.getLogger().setLevel(logging.DEBUG)
    requests_log = logging.getLogger("requests.packages.urllib3")
    requests_log.setLevel(logging.DEBUG)
    requests_log.propagate = True



#
# Copyright (c) 2015-2018 Canonical, Ltd.
#
# This file is part of Talisker
# (see http://github.com/canonical-ols/talisker).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import *  # noqa
import logging
import sys
import textwrap
import os

from future.utils import exec_
# be *very* careful about what is imported here, as any stdlib loggers that are
# created can not be changed!
from talisker.config import get_config
from talisker.util import ensure_extra_versions_supported, pkg_is_installed
from talisker.context import CONTEXT, clear as clear_context  # noqa

__version__ = '0.11.0'
__all__ = [
    'initialise',
    'get_config',
    'run',
    'run_gunicorn',
    'run_celery',
    'run_gunicorn_eventlet',
    'run_gunicorn_gevent',
]
_early_log_messages = []
prometheus_multiproc_cleanup = False


def _log(name, level, *args, **kwargs):
    """Deferred log wrapper"""
    logger = logging.getLogger(name)
    getattr(logger, level)(*args, **kwargs)


def early_log(name, level, *args, **kwargs):
    """Logger wrap for talisker startup code.

    Collects logs for later processing when logging is initialised
    """
    _early_log_messages.append((name, level, args, kwargs))


def _flush_early_logs():
    global early_log
    # process pending logs
    for name, level, args, kwargs in _early_log_messages:
        _log(name, level, *args, **kwargs)
    _early_log_messages[:] = []

    # switch to immediate logging for any further early logs
    early_log = _log


def initialise(env=os.environ):
    global early_log

    config = get_config(env)
    import talisker.logs
    talisker.logs.configure(config)
    _flush_early_logs()

    # now that logging is set up, initialise other modules
    # sentry first, so we can report any further errors in initialisation
    # TODO: add deferred logging, so we can set up sentry first thing
    import talisker.sentry
    talisker.sentry.get_client()
    import talisker.statsd
    talisker.statsd.get_client()
    clear_contexts()
    return config


def clear_contexts():
    """Helper to clear any thread local contexts."""
    import talisker.sentry
    clear_context()
    client = talisker.sentry.get_client()
    client.context.clear()
    client.transaction.clear()


class RunException(Exception):
    pass


def run():
    """Initialise Talisker then exec python script."""
    initialise()
    logger = logging.getLogger('talisker.run')

    name = sys.argv[0]
    if '__main__.py' in name:
        # friendlier message
        name = '{} -m talisker'.format(sys.executable)

    extra = {}
    try:
        if len(sys.argv) < 2:
            raise RunException('usage: {} <script>  ...'.format(name))

        script = sys.argv[1]
        extra['script'] = script
        with open(script, 'rb') as f:
            code = compile(f.read(), script, 'exec')

        # pretend we just invoked python script.py by mimicing usual python
        # behavior
        sys.path.insert(0, os.path.dirname(script))
        sys.argv = sys.argv[1:]
        globs = {}
        globs['__file__'] = script
        globs['__name__'] = '__main__'
        globs['__package__'] = None

        clear_contexts()
        exec_(code, globs, None)

    except Exception:
        logger.exception('Unhandled exception', extra=extra)
        sys.exit(1)


def format_docstring(docstring, width):
    short, _, long = docstring.partition('\n\n')
    short = textwrap.wrap(' '.join(short.split()), width)
    if long:
        long = textwrap.wrap(' '.join(long.split()), width)
    return short, long


def run_help():
    """
    Usage: talisker.help [CONFIG NAME]

    Talisker provides some executable wrappers, which initialise Talisker
    and then simply pass through any supplied arguments to underlying command.

     - talisker.gunicorn wraps the regular gunicorn invocation
     - talisker.celery wraps the celery command
     - talisker.run wraps a regular call to python, and takes a script to run

    Talisker can be configured by the environment variables listed below. These
    variable can also be supplied in a python file, although environment
    variables override any file configuration.
    """
    width = 80
    indent = 30
    rest = width - indent
    indent_str = '\n' + ' ' * indent
    metadata = get_config().metadata()

    if len(sys.argv) > 1:
        name = sys.argv[1]
        if name.upper() not in metadata:
            sys.stderr.write('Invalid config: {}\n'.format(name))
            sys.exit('Invalid config: {}'.format(name))

        doc = metadata[name].doc
        if doc is None:
            short = 'No documentation'
            long = []
        else:
            short, long = format_docstring(doc, width)

        print(name)
        print()
        print('\n'.join(short))
        print()
        if long:
            print('\n'.join(long))
            print()

    else:
        # print header
        print(textwrap.dedent(run_help.__doc__).lstrip())
        print()

        for name, meta in metadata.items():
            if meta.doc is not None:
                short, long = format_docstring(meta.doc, rest)
                print('{:{indent}}{}'.format(
                    name, indent_str.join(short), indent=indent)
                )


def run_celery(argv=sys.argv):
    initialise()
    os.environ['CELERYD_REDIRECT_STDOUTS'] = 'False'
    # technically we don't need this, as we disable celery's logging
    # altogether, but it doesn't hurt
    os.environ['CELERYD_HIJACK_ROOT_LOGGER'] = 'False'
    ensure_extra_versions_supported('celery')
    import talisker.celery
    from celery.bin.celery import main
    talisker.celery.enable_signals()
    clear_contexts()
    main(argv)


def run_gunicorn():
    config = get_config()

    # configure prometheus_client early as possible
    if pkg_is_installed('prometheus-client'):
        # Early throw-away parsing of gunicorn config, as we need to decide
        # whether to enable prometheus multiprocess before we start importing
        from gunicorn.app.wsgiapp import WSGIApplication
        g_cfg = WSGIApplication().cfg
        if g_cfg.workers > 1 or 'prometheus_multiproc_dir' in os.environ:
            from talisker.prometheus import setup_prometheus_multiproc
            async_workers = ('gevent', 'eventlet')
            # must be done before prometheus_client is imported *anywhere*
            setup_prometheus_multiproc(
                any(n in g_cfg.worker_class_str for n in async_workers)
            )

    initialise()

    import talisker.gunicorn

    if pkg_is_installed('celery'):
        import talisker.celery
        talisker.celery.enable_signals()

    app = talisker.gunicorn.TaliskerApplication(
        "%(prog)s [OPTIONS] [APP_MODULE]", config.devel, config.debuglog)
    clear_contexts()
    return app.run()


# these two entrypoints workaround a bug in requests on python 3.6 for tls
# https://github.com/requests/requests/issues/3752
# they may go away once requests fixes this issue
def run_gunicorn_eventlet():
    # this is taken from gunicorn EventletWorker.patch()
    import eventlet
    eventlet.monkey_patch(os=False)
    run_gunicorn()


def run_gunicorn_gevent():
    import gevent
    # this is taken from gunicorn GeventWorker.patch()
    from gevent import monkey
    if gevent.version_info[0] == 0:
        monkey.patch_all()
    else:
        monkey.patch_all(subprocess=True)
    run_gunicorn()



#
# Copyright (c) 2015-2018 Canonical, Ltd.
#
# This file is part of Talisker
# (see http://github.com/canonical-ols/talisker).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import *  # noqa
__metaclass__ = type

import ast
import functools
import logging
import json
import os
import re
import shlex
import subprocess
import sys
import tempfile
import time
import uuid
import zlib

import raven
import requests

import talisker.context
import talisker.logs
import talisker.requests
import talisker.statsd
import talisker.util


__all__ = [
    'configure_testing',
    'clear_all'
    'TestContext',
]

HAVE_DJANGO_INSTALLED = talisker.util.pkg_is_installed('django')

if sys.version_info[0] == 2:
    def temp_file():
        return tempfile.NamedTemporaryFile('wb', bufsize=0)
else:
    def temp_file():
        return tempfile.NamedTemporaryFile('wb', buffering=0)


def clear_all():
    """Clear all talisker state."""
    talisker.context.clear()  # talisker request_context
    talisker.requests.clear()  # talisker requests.Session cache
    talisker.sentry.clear()  # sentry per-request state
    talisker.util.clear_globals()  # module caches
    talisker.util.clear_context_locals()  # any remaining contexts
    clear_sentry_messages()


def configure_testing():
    """Set up a null handler for logging and testing sentry remote."""
    talisker.logs.configure_test_logging()
    configure_sentry_client()


class LogRecordList(list):
    """A container for searching a list of logging.LogRecords."""
    _sentinel = object()

    def _clean_kwargs(self, kwargs):
        # some UX tweaks
        # shortcut for level, so we can do e.g. level=logging.INFO,
        level = kwargs.pop('level', None)
        if level is not None:
            if isinstance(level, str):
                kwargs['levelname'] = level.upper()
            else:
                kwargs['levelno'] = level
        # you can case levelname either way
        elif 'levelname' in kwargs:
            kwargs['levelname'] = kwargs['levelname'].upper()

    def _match(self, record, extra, kwargs):
        _s = self._sentinel

        def cmp(a, b):
            if a == b:
                return True
            # partial match for strings
            try:
                return b in a
            except Exception:
                return False

        if all(cmp(getattr(record, k, _s), v) for k, v in kwargs.items()):
            if extra:
                get = record.extra.get
                if all(cmp(get(k, _s), v) for k, v in extra.items()):
                    return True
            else:
                return True
        else:
            return False

    def filter(self, extra=None, **kwargs):
        """Search for records matching the query parameters.

        The query parameters are attributes of the logging.LogRecord instance,
        or also the generic 'level' parameter, as shorthand for
        record.levelname or record.levelno. The `extra` dict is compared
        against the LogRecord's extra dict also.

        Matching is partial for strings (i.e. if a in b).
        """
        self._clean_kwargs(kwargs)
        found = self.__class__()
        for record in self:
            if self._match(record, extra, kwargs):
                found.append(record)
        return found

    def find(self, extra=None, **kwargs):
        """Return the first record that matches the query parameters.

        Accepts the same parameters as filter()"""
        self._clean_kwargs(kwargs)
        for record in self:
            if self._match(record, extra, kwargs):
                return record

    def exists(self, extra=None, **kwargs):
        """Return true if the matching log message exists."""
        return self.find(extra, **kwargs) is not None

    def match(self, record, extra=None, **kwargs):
        """Match an single record, same as parameters to filter()."""
        self._clean_kwargs(kwargs)
        return self._match(record, extra, kwargs)


class TestHandler(logging.Handler):
    """Testing handler that records its logs in memory."""
    def __init__(self, level=logging.NOTSET):
        super().__init__(level)
        self.records = LogRecordList()
        self.lines = []

    def emit(self, record):
        self.records.append(record)
        # formatting forces the setting of record.message
        self.lines.extend(self.format(record).split('\n'))


class DummySentryTransport(raven.transport.Transport):
    """Fake sentry transport for testing."""
    scheme = ['test']

    def __init__(self, *args, **kwargs):
        self.messages = []

    def send(self, *args, **kwargs):
        # In raven<6, args = (data, headers).
        # In raven 6.x args = (url, data, headers)
        if len(args) == 2:
            data, _ = args
        elif len(args) == 3:
            _, data, _ = args
        else:
            raise Exception('raven Transport.send api seems to have changed')
        raw = json.loads(zlib.decompress(data).decode('utf8'))
        # to make asserting easier, parse json strings into python strings
        for k, v in list(raw['extra'].items()):
            try:
                val = ast.literal_eval(v)
            except Exception:
                pass
            else:
                raw['extra'][k] = val

        self.messages.append(raw)


def clear_sentry_messages():
    messages = get_sentry_messages()
    if messages is not None:
        # py2.7 doesn't have list.clear() :(
        messages[:] = []


def get_sentry_messages(client=None):
    if client is None:
        client = talisker.sentry.get_client()
    transport = client.remote.get_transport()
    if transport is None:
        return None
    else:
        return transport.messages


TEST_SENTRY_DSN = 'http://user:pass@host/project'


def configure_sentry_client(client=None):
    client = talisker.sentry.get_client()
    client.set_dsn(TEST_SENTRY_DSN, transport=DummySentryTransport)


class TestContext():

    def __init__(self, name=None):
        if name is None:
            self.name = str(uuid.uuid4())
        else:
            self.name = name

        self.dsn = TEST_SENTRY_DSN + self.name
        self.handler = TestHandler()
        self.statsd_client = talisker.statsd.DummyClient(collect=True)
        self.sentry_client = talisker.sentry.get_client()
        self.sentry_remote = self.sentry_client.remote

    def start(self):
        self.old_statsd = talisker.statsd.get_client.raw_update(
            self.statsd_client)
        self.sentry_client.set_dsn(self.dsn, transport=DummySentryTransport)
        self.sentry_transport = self.sentry_client.remote.get_transport()
        talisker.logs.add_talisker_handler(logging.NOTSET, self.handler)

    def stop(self):
        logging.getLogger().handlers.remove(self.handler)
        talisker.statsd.get_client.raw_update(self.old_statsd)
        # restore the original sentry remote
        self.sentry_client.remote = self.sentry_remote
        self.sentry_client._transport_cache.pop(self.dsn, None)
        clear_all()

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type=None, exc_value=None, exc_traceback=None):
        self.stop()

    def __call__(self, func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            with self:
                return func(*args, **kwargs)
        return wrapper

    @property
    def logs(self):
        return self.handler.records

    @property
    def sentry(self):
        return self.sentry_transport.messages

    @property
    def statsd(self):
        return self.statsd_client.stats

    def assert_log(self, **kwargs):
        if not self.logs.exists(**kwargs):
            # evaluate each term independently to narrow down culprit
            terms = []
            for kw, value in kwargs.items():
                num = len(self.logs.filter(**{kw: value}))
                terms.append((num, kw, value))
            terms.sort()  # 0 matches go first, as likely to be the issue

            desc = '\n    '.join(
                '{1}={2!r} ({0} matches)'.format(*t) for t in terms
            )
            raise AssertionError(
                'Could not find log out of {} logs:\n    {}'.format(
                    len(self.logs), desc))

    def assert_not_log(self, **kwargs):
        if self.logs.exists(**kwargs):
            desc = '\n    '.join(
                '{}={!r}'.format(k, v) for k, v in sorted(kwargs.items())
            )
            raise AssertionError(
                'Found log matching the following:\n    {}'.format(desc)
            )


class LogOutput:
    """A container for log messages output by a Talisker program.

    Parses log lines, and makes them available for inspection in tests.
    """

    TIMESTAMP = re.compile(r'^\d\d\d\d-\d\d-\d\d \d\d:\d\d:\d\d.\d\d\dZ')

    def __init__(self, lines):
        self.logs = list(self.read(lines))

    def read(self, lines):
        current = []
        for line in lines:
            if line.strip():
                if self.TIMESTAMP.match(line):
                    if current:
                        yield self.parse(current)
                    current = []
                current.append(line)
        if current:
            yield self.parse(current)

    def parse(self, logs):
        """Stupid simple logfmt parser"""
        log = logs[0]
        trailer = logs[1:]
        parsed = shlex.split(log)
        try:
            date, time, level, name, msg = parsed[:5]
            extra = dict((v.split('=', 1)) for v in parsed[5:])
        except ValueError:
            assert 0, "failed to parse logfmt:\n" + '\n'.join(logs)
        return {
            'ts': date + " " + time,
            'level': level,
            'logger': name,
            'logmsg': msg,
            'extra': extra,
            'trailer': trailer,
        }

    def _compare_strings(self, needle, haystack):
        return (
            all(k in haystack for k in needle)
            and all(needle[k] in haystack[k] for k in needle)
        )

    def exists(self, **match):
        extra = match.pop('extra', None)
        trailer = match.pop('trailer', None)
        for log in self.logs:
            strings_match = self._compare_strings(match, log)
            extra_match = trailer_match = True

            if extra is not None:
                extra_match = self._compare_strings(extra, log['extra'])

            if trailer is not None:
                for line in trailer:
                    if not any(line in tline for tline in log['trailer']):
                        trailer_match = False
                        break

            if strings_match and extra_match and trailer_match:
                return True

        return False

    def __contains__(self, params):
        return self.exists(**params)


class ServerProcessError(Exception):
    pass


class ServerProcess(object):
    """Context mananger to run a server subprocess """
    def __init__(self, cmd, env=None, **kwargs):
        self.cmd = cmd
        self.env = env
        self.kwargs = kwargs
        self.output = []
        self.ps = None
        self._log = None

    def start(self):
        self.output_file = temp_file()
        self.reader = open(self.output_file.name, 'r')

        if self.env is None:
            env = os.environ.copy()
        else:
            env = self.env.copy()

        if 'PYTHONUNBUFFERED' not in env:
            env['PYTHONUNBUFFERED'] = '1'

        self.ps = subprocess.Popen(
            self.cmd,
            bufsize=0,
            stdout=self.output_file,
            stderr=subprocess.STDOUT,
            stdin=subprocess.PIPE,
            universal_newlines=True,
            env=env,
            **self.kwargs
        )
        try:
            self.check()
        except Exception:
            self.close(error=True)
            raise

    def close(self, error=False):
        """Clean up process."""
        if not self.finished:
            self.ps.terminate()
            self.ps.wait()

        if not self.reader.closed:
            for line in self.reader.readlines():
                self.output.append(line.rstrip())
            self.reader.close()

        if error:
            # just dump the output to stderr for now, for visibility
            sys.stderr.write('Server process died:\n')
            sys.stderr.write('\n'.join(self.output))

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type=None, exc_value=None, exc_traceback=None):
        self.close(error=exc_type is not None)

    @property
    def finished(self):
        return self.ps is not None and self.ps.poll() is not None

    @property
    def log(self):
        assert self.finished
        if self._log is None:
            self._log = LogOutput(self.output)
        return self._log

    def check(self):
        """Check the server process is still running or has finished."""
        rc = self.ps.poll()
        if rc is not None and rc > 0:
            raise ServerProcessError('subprocess errored')

    def readline(self, timeout=30, delay=0.1):
        if not self.output_file.closed:
            self.output_file.flush()

        start = time.time()
        line = self.reader.readline()

        while time.time() - start < timeout and line == '':
            self.check()
            time.sleep(delay)
            line = self.reader.readline()

        if line == '':
            raise Exception(
                'could not read line from process stdout '
                'within timeout of {}'.format(timeout))

        self.output.append(line.rstrip())
        return timeout - (time.time() - start)

    def wait_for_output(self, target, timeout, delay=0.1):
        try:
            # read first line if needed
            if len(self.output) == 0:
                self.readline(timeout, delay)

            # maybe we already got there
            for line in self.output:
                if target in line:
                    return

            while target not in self.output[-1]:
                self.readline(timeout, delay)

        except Exception:
            self.close(error=True)
            raise


class GunicornProcess(ServerProcess):
    """Context mananger to run Talisker's gunicorn server.

    It captures all output, and waits untill it sees which port gunicorn is
    listening on, expose as url attribute.
    """

    ADDRESS = re.compile(r'http://127\.0\.0\.1:(\d+)')
    WORKER = 'Booting worker with pid'

    def __init__(self,
                 app,
                 args=None,
                 env=None,
                 gunicorn='talisker.gunicorn',
                 ip='127.0.0.1',
                 **kwargs):

        self.app = app
        self.ip = ip
        self.port = None
        cmd = [
            gunicorn,
            '--bind', ip + ':0',
            '--access-logfile', '-',
        ]
        if args:
            cmd.extend(args)
        cmd.append(app)
        super().__init__(cmd, env=env, **kwargs)

    def start(self):
        super().start()

        self.wait_for_output(self.WORKER, timeout=30)
        for line in self.output:
            m = self.ADDRESS.search(line)
            if m:
                self.port = m.groups()[0]

        if self.port is None:
            raise Exception('could not parse gunicorn port from output')

        # check that the app has loaded and gunicorn has not died before
        # returning control.
        try:
            self.ping()
        except Exception:
            self.close(error=True)
            raise

    def url(self, path):
        return 'http://{}:{}{}'.format(self.ip, self.port, path)

    def ping(self):
        success = False
        try:
            r = requests.get(self.url('/_status/ping'))
            success = r.status_code == 200
        except requests.exceptions.RequestException:
            pass

        if not success:
            raise ServerProcessError(
                'could not ping server at {}'.format(self.url('/')),
            )


def run_wsgi(app, environ):
    """Execute a wsgi application, returning (body, status, headers)."""
    output = {}

    def start_response(status, headers, exc_info=None):
        output['status'] = status
        output['headers'] = headers

    body = app(environ, start_response)

    return body, output['status'], output['headers']



#
# Copyright (c) 2015-2018 Canonical, Ltd.
#
# This file is part of Talisker
# (see http://github.com/canonical-ols/talisker).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import *  # noqa
import logging
from raven.contrib.django.client import DjangoClient

import talisker.sentry


# raven's django support does some very odd things.
# There is a module global, that is a proxy to the output of
# raven.contrib.django.models.get_client()
# But the only way you can customise the set up is via subclassing the client.
# So that's what we do. We ensure talisker's configuration, and hook it in to
# the other things that need to know about the client.  Django users just need
# to add the following to settings:
# SENTRY_CLIENT = 'talisker.django.SentryClient'

class SentryClient(DjangoClient):
    def __init__(self, *args, **kwargs):
        # SQL hook sends raw SQL to the server. Not cool, bro.
        kwargs['install_sql_hook'] = False
        talisker.sentry.ensure_talisker_config(kwargs)
        logging.getLogger(__name__).info(
            'updating raven config from django app')
        super().__init__(*args, **kwargs)
        # update any previously configured sentry client
        talisker.sentry.set_client(self)

    def build_msg(self, event_type, *args, **kwargs):
        data = super().build_msg(event_type, *args, **kwargs)
        talisker.sentry.add_talisker_context(data)
        return data

    def set_dsn(self, dsn=None, transport=None):
        super().set_dsn(dsn, transport)
        talisker.sentry.log_client(self)


def middleware(get_response):
    """Set up middleware to add X-View-Name header."""
    def add_view_name(request):
        response = get_response(request)
        if getattr(request, 'resolver_match', None):
            response['X-View-Name'] = request.resolver_match.view_name
        return response
    return add_view_name



#
# Copyright (c) 2015-2018 Canonical, Ltd.
#
# This file is part of Talisker
# (see http://github.com/canonical-ols/talisker).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import *  # noqa
__metaclass__ = type

try:
    from collections.abc import Mapping
except ImportError:  # < py3.3
    from collections import Mapping

from collections import OrderedDict, defaultdict
from contextlib import contextmanager

from werkzeug.local import release_local
from talisker.util import context_local

# a per request/job context. Generally, this will be the equivalent of thread
# local storage, but if greenlets are being used it will be a greenlet local.
CONTEXT = context_local()


def clear():
    release_local(CONTEXT)


class ContextStack(Mapping):
    """A stacked set of dicts stored in a context.

    Support lookups and iteration, which go from top of the stack down.
    Can also be used as a context manager.

    """
    def __init__(self, name, *dicts):
        """Initialise stack, with name to use in context storage."""
        self._name = name
        self._stack.extend(dicts)

    def __eq__(self, other):
        return (
            self._name == other._name
            and list(self) == list(other)
        )

    def _clear(self):
        storage = {'stack': [], 'flattened': None}
        setattr(CONTEXT, self._name, storage)
        return storage

    def clear(self):
        """Clear the stack."""
        self._clear()

    @property
    def _storage(self):
        storage = getattr(CONTEXT, self._name, None)
        if storage is None:
            storage = self._clear()
        return storage

    @property
    def _stack(self):
        return self._storage['stack']

    @property
    def flat(self):
        """Cached flattened dict"""
        _flat = self._storage.get('flattened', None)
        if _flat is None:
            _flat = self._storage['flattened'] = OrderedDict(self._iterate())
        return _flat

    def _clear_flat(self):
        self._storage['flattened'] = None

    def _iterate(self):
        """Iterate from top to bottom, preserving individual dict ordering."""
        seen = set()
        for d in reversed(self._stack):
            for k, v in d.items():
                if k not in seen:
                    yield k, v
            seen = seen.union(d)

    def push(self, _dict=None, **kwargs):
        """Add a new dict to the stack.

        Can take a single positional argument, which is a dict, and/or kwargs
        dict to use.

        Returns the stack level before adding this dict, for use with
        unwind."""
        if _dict is None:
            d = {}
        else:
            d = _dict.copy()
        d.update(kwargs)
        level = len(self._stack)
        self._stack.append(d)
        self._clear_flat()
        return level

    def pop(self):
        """Pop the most recent dict from the stack"""
        if self._stack:
            self._stack.pop()
        self._clear_flat()

    def unwind(self, level):
        """Unwind the stack to a specific level."""
        while len(self._stack) > level:
            self._stack.pop()
        self._clear_flat()

    @contextmanager
    def __call__(self, extra=None, **kwargs):
        """Context manager to push/run/pop."""
        self.push(extra, **kwargs)
        yield self
        self.pop()

    def __getitem__(self, item):
        """Key lookup, from top to bottom."""
        return self.flat[item]

    def __len__(self):
        return len(self.flat)

    def __iter__(self):
        """Iterate from top to bottom, preserving individual dict ordering."""
        return iter(self.flat)


class Tracker():
    def __init__(self):
        self.count = 0
        self.time = 0.0


def track_request_metric(type, duration):
    tracking = getattr(CONTEXT, 'request_tracking', defaultdict(Tracker))
    tracking[type].count += 1
    tracking[type].time += duration
    CONTEXT.request_tracking = tracking



#
# Copyright (c) 2015-2018 Canonical, Ltd.
#
# This file is part of Talisker
# (see http://github.com/canonical-ols/talisker).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import *  # noqa
__metaclass__ = type

from contextlib import contextmanager
from future.moves.urllib.parse import urlparse, parse_qs

from statsd import defaults
from statsd.client import StatsClient

import talisker
from talisker.util import module_cache

__all__ = ['get_client']


def parse_statsd_dsn(dsn):
    parsed = urlparse(dsn)
    host = parsed.hostname
    port = parsed.port or defaults.PORT
    prefix = None
    if parsed.path:
        prefix = parsed.path.strip('/').replace('/', '.')
    ipv6 = parsed.scheme in ('udp6', 'tcp6')
    size = int(
        parse_qs(parsed.query).get('maxudpsize', [defaults.MAXUDPSIZE])[0])
    return host, port, prefix, size, ipv6


@module_cache
def get_client():
    client = None
    dsn = talisker.get_config().statsd_dsn
    if dsn is None:
        client = DummyClient()
    else:
        if not dsn.startswith('udp'):
            raise Exception('Talisker only supports udp stastd client')
        client = StatsClient(*parse_statsd_dsn(dsn))

    return client


class DummyClient(StatsClient):  # lgtm [py/missing-call-to-init]
    """Mock client for statsd that can collect data when testing."""
    _prefix = ''  # force no prefix

    def __init__(self, collect=False):
        # Note: do *not* call super(), as that will create udp socket we don't
        # want.
        if collect:
            self.stats = MetricList()
        else:
            self.stats = None

    def _send(self, data):  # lgtm [py/inheritance/signature-mismatch]
        if self.stats is not None:
            self.stats.append(data)

    def pipeline(self):
        return self.__class__(collect=True)

    # pipeline methods
    def send(self):
        if self.stats:
            self.stats[:] = []

    def __enter__(self):
        return self

    def __exit__(self, typ, value, tb):
        self.send()

    # test helper methods
    @contextmanager
    def collect(self):
        orig_stats = self.stats
        self.stats = []
        yield self.stats
        self.stats = orig_stats


class MetricList(list):
    """A container for searching a list of statsd metrics."""

    def filter(self, name):
        filtered = self.__class__()
        for metric in self:
            if name in metric:
                filtered.append(metric)
        return filtered



#
# Copyright (c) 2015-2018 Canonical, Ltd.
#
# This file is part of Talisker
# (see http://github.com/canonical-ols/talisker).
#
# Licensed to the Apache Software Foundation (ASF) under one
# or more contributor license agreements.  See the NOTICE file
# distributed with this work for additional information
# regarding copyright ownership.  The ASF licenses this file
# to you under the Apache License, Version 2.0 (the
# "License"); you may not use this file except in compliance
# with the License.  You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing,
# software distributed under the License is distributed on an
# "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY
# KIND, either express or implied.  See the License for the
# specific language governing permissions and limitations
# under the License.
#

# -*- coding: utf-8 -*-
from __future__ import unicode_literals
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import *  # noqa
from collections import defaultdict, OrderedDict
from contextlib import contextmanager
from multiprocessing import Lock
import errno
import json
import logging
import os
import tempfile
import time

import talisker
from talisker.util import pkg_is_installed, TaliskerVersionException


prometheus_installed = pkg_is_installed('prometheus_client')
if prometheus_installed and prometheus_installed.version in ('0.4.0', '0.4.1'):
    raise TaliskerVersionException(
        'prometheus_client {} has a critical bug in multiprocess mode, '
        'and is not supported in Talisker. '
        'https://github.com/prometheus/client_python/issues/322'.format(
            prometheus_installed.version,
        )
    )


_lock = None
histogram_archive = 'histogram_archive.db'
counter_archive = 'counter_archive.db'


class PrometheusLockTimeout(Exception):
    pass


@contextmanager
def try_prometheus_lock_noop(timeout=10.0):
    """Default implementation: does nothing."""
    yield


try_prometheus_lock = try_prometheus_lock_noop


@contextmanager
def try_prometheus_lock_normal(timeout=10.0):
    """Try acquire the multiprocess lock, with timeout.

    Note: the timeout is implemented in C, so will block in async contexts.
    """
    if not _lock.acquire(timeout=timeout):
        raise PrometheusLockTimeout('Timeout acquiring prometheus lock')
    yield
    _lock.release()


@contextmanager
def try_prometheus_lock_patched_async(timeout=10.0):
    """Try acquire the multiprocss lock, with timeout in python code.

    This puts the timeout into a loop in application python code, so it does
    not block when using gevent/eventlet monkey patched workers.
    """
    deadline = time.time() + timeout
    while time.time() < deadline:
        if _lock.acquire(block=False):
            break
        else:
            time.sleep(0.1)
    else:
        raise PrometheusLockTimeout('Timeout acquiring prometheus lock')

    yield
    _lock.release()


def setup_prometheus_multiproc(async_mode):
    """Setup prometheus_client multiprocess support.

    This involves setting up locking and a temporary directory if needed.

    Note: if we can't create the lock (e.g. inside a strictly confined snap
    package) then multiprocess cleanup is *not* enabled.
    """
    global _lock, registry, try_prometheus_lock
    if not prometheus_installed:
        return

    if 'prometheus_multiproc_dir' not in os.environ:
        prefix = 'prometheus_multiproc_{}_'.format(os.getpid())
        tmp = tempfile.mkdtemp(prefix=prefix)
        os.environ['prometheus_multiproc_dir'] = tmp

    # try enable multiprocess cleanup
    try:
        _lock = Lock()
    except OSError as exc:
        if exc.errno != errno.EACCES:
            raise

        talisker.early_log(
            __name__,
            'warning',
            'Unable to create lock for prometheus, cleanup disabled',
        )
    else:

        if async_mode:
            try_prometheus_lock = try_prometheus_lock_patched_async
        else:
            try_prometheus_lock = try_prometheus_lock_normal

        # signal to others that clean up is enabled
        talisker.prometheus_multiproc_cleanup = True

    talisker.early_log(
        __name__,
        'info',
        'prometheus_client is in multiprocess mode',
        extra={
            'multiproc_dir': os.environ['prometheus_multiproc_dir'],
            'cleanup_enabled': talisker.prometheus_multiproc_cleanup,
        }
    )


def collect_metrics():
    from prometheus_client import (
        CollectorRegistry,
        core,
        generate_latest,
        multiprocess,
    )
    if 'prometheus_multiproc_dir' in os.environ:
        registry = CollectorRegistry()
        multiprocess.MultiProcessCollector(registry)
    else:
        registry = core.REGISTRY
    with try_prometheus_lock():
        return generate_latest(registry)


def _filter_exists(paths):
    exists = []
    for path in paths:
        if os.path.exists(path):
            exists.append(path)
    return exists


def prometheus_cleanup_worker(pid):
    """Aggregate dead worker's metrics into a single archive file."""
    from prometheus_client import multiprocess
    multiprocess.mark_process_dead(pid)  # this takes care of gauges
    prom_dir = os.environ['prometheus_multiproc_dir']
    worker_files = [
        'histogram_{}.db'.format(pid),
        'counter_{}.db'.format(pid),
    ]
    paths = _filter_exists(os.path.join(prom_dir, f) for f in worker_files)

    # check at least one worker file exists
    if not paths:
        return

    histogram_path = os.path.join(prom_dir, histogram_archive)
    counter_path = os.path.join(prom_dir, counter_archive)
    archive_paths = _filter_exists([histogram_path, counter_path])

    collect_paths = paths + archive_paths
    collector = multiprocess.MultiProcessCollector(None)

    try:
        metrics = collector.merge(collect_paths, accumulate=False)
    except AttributeError:
        metrics = legacy_collect(collect_paths)

    tmp_histogram = tempfile.NamedTemporaryFile(delete=False)
    tmp_counter = tempfile.NamedTemporaryFile(delete=False)
    write_metrics(metrics, tmp_histogram.name, tmp_counter.name)

    try:
        # ensure reader does get partial state
        with try_prometheus_lock():
            os.rename(tmp_histogram.name, histogram_path)
            os.rename(tmp_counter.name, counter_path)

            for path in paths:
                os.unlink(path)
    except PrometheusLockTimeout:
        logging.getLogger(__name__).exception(
            'Failed to acquire prometheus lock to clean worker files',
            extra={
                'pid': pid,
                'paths': paths,
            }
        )


def legacy_collect(files):
    """This almost verbatim from MultiProcessCollector.collect(), pre 0.4.0

    The original collects all results in a format designed to be scraped. We
    instead need to collect limited results, in a format that can be written
    back to disk. To facilitate this, this version of collect() preserves label
    ordering, and does not aggregate the histograms.

    Specifically, it differs from the original:

    1. it takes its files as an argument, rather than hardcoding '*.db'
    2. it does not accumulate histograms
    3. it preserves label order, to facilitate being inserted back into an mmap
       file.

    It needs to be kept up to date with changes to prometheus_client as much as
    possible, or until changes are landed upstream to allow reuse of collect().
    """
    from prometheus_client import core
    metrics = {}
    for f in files:
        if not os.path.exists(f):
            continue
        # verbatim from here...
        parts = os.path.basename(f).split('_')
        typ = parts[0]
        d = core._MmapedDict(f, read_mode=True)
        for key, value in d.read_all_values():
            # Note: key format changed in 0.4+
            metric_name, name, labelnames, labelvalues = json.loads(key)

            metric = metrics.get(metric_name)
            if metric is None:
                metric = core.Metric(metric_name, 'Multiprocess metric', typ)
                metrics[metric_name] = metric

            if typ == 'gauge':
                pid = parts[2][:-3]
                metric._multiprocess_mode = parts[1]
                metric.add_sample(
                    name,
                    tuple(zip(labelnames, labelvalues)) + (('pid', pid), ),
                    value,
                )
            else:
                # The duplicates and labels are fixed in the next for.
                metric.add_sample(
                    name,
                    tuple(zip(labelnames, labelvalues)),
                    value,
                )
        d.close()

    for metric in metrics.values():
        samples = defaultdict(float)
        buckets = {}
        for name, labels, value in metric.samples:
            if metric.type == 'gauge':
                without_pid = tuple(l for l in labels if l[0] != 'pid')
                if metric._multiprocess_mode == 'min':
                    current = samples.setdefault((name, without_pid), value)
                    if value < current:
                        samples[(name, without_pid)] = value
                elif metric._multiprocess_mode == 'max':
                    current = samples.setdefault((name, without_pid), value)
                    if value > current:
                        samples[(name, without_pid)] = value
                elif metric._multiprocess_mode == 'livesum':
                    samples[(name, without_pid)] += value
                else:  # all/liveall
                    samples[(name, labels)] = value

            elif metric.type == 'histogram':
                bucket = tuple(float(l[1]) for l in labels if l[0] == 'le')
                if bucket:
                    # _bucket
                    without_le = tuple(l for l in labels if l[0] != 'le')
                    buckets.setdefault(without_le, {})
                    buckets[without_le].setdefault(bucket[0], 0.0)
                    buckets[without_le][bucket[0]] += value
                else:
                    # _sum/_count
                    samples[(name, labels)] += value

            else:
                # Counter and Summary.
                samples[(name, labels)] += value

        # end of verbatim copy
        # modified to remove accumulation
        if metric.type == 'histogram':
            for labels, values in buckets.items():
                for bucket, value in sorted(values.items()):
                    key = (
                        metric.name + '_bucket',
                        labels + (('le', core._floatToGoString(bucket)),),
                    )
                    samples[key] = value

        # Convert to correct sample format.
        metric.samples = [
            # OrderedDict used instead of dict
            (name, OrderedDict(labels), value)
            for (name, labels), value in samples.items()
        ]
    return metrics.values()


def write_metrics(metrics, histogram_file, counter_file):
    from prometheus_client import core
    try:
        key_func = core._mmap_key
    except AttributeError:
        # pre 0.4 key format
        def key_func(metric_name, name, labelnames, labelvalues):
            return json.dumps(
                (metric_name, name, tuple(labels), tuple(labels.values()))
            )

    histograms = core._MmapedDict(histogram_file)
    counters = core._MmapedDict(counter_file)

    try:
        for metric in metrics:
            if metric.type == 'histogram':
                sink = histograms
            elif metric.type == 'counter':
                sink = counters
            else:
                continue

            for sample in metric.samples:
                # prometheus_client 0.4+ adds extra fields
                name, labels, value = sample[:3]
                key = key_func(
                    metric.name,
                    name,
                    tuple(labels),
                    tuple(labels.values()),
                )
                sink.write_value(key, value)
    finally:
        histograms.close()
        counters.close()
